"""
Lightweight telemetry wrapper for svVascularize.

This module provides optional integration with Sentry for crash and error
reporting, primarily for the GUI. It is designed to be safe when Sentry is
not installed or when no DSN is configured.
"""
from __future__ import annotations

import os
import sys
from typing import Optional

_ENABLED: bool = False
_EXCEPTHOOK_INSTALLED: bool = False

# Default DSN for the svVascularize GUI. This is used only when no explicit
# DSN is provided by the caller or via environment variables. It is safe for
# Sentry DSNs to be embedded in client applications, but environment variables
# still take precedence when present.
_DEFAULT_GUI_DSN: str = (
    "https://c76b4f3a871ce81fb7d5b55df436e6fc@"
    "o4510450670174208.ingest.us.sentry.io/4510450824577024"
)


def _before_send(event, hint):
    """
    Scrub potentially sensitive fields before sending to Sentry.

    This removes obvious user/request payloads. Additional scrubbing can be
    added here if needed (e.g., trimming filesystem paths).
    """
    try:
        event.pop("user", None)
        event.pop("request", None)
    except Exception:
        pass
    return event


def init_telemetry(dsn: Optional[str], release: str, environment: str = "gui") -> None:
    """
    Initialize Sentry-based telemetry if a DSN is provided.

    Parameters
    ----------
    dsn : str | None
        Sentry DSN. When ``None`` or empty, telemetry is disabled.
    release : str
        Application/library release string (e.g., svv.__version__).
    environment : str
        Environment label (e.g., "gui", "dev", "staging").
    """
    global _ENABLED, _EXCEPTHOOK_INSTALLED

    if _ENABLED:
        return

    # Allow environment variable override to hard-disable telemetry
    if os.environ.get("SVV_TELEMETRY_DISABLED", "").strip() == "1":
        return

    # Use provided DSN when available; otherwise fall back to the built-in
    # default DSN for the svVascularize GUI.
    if not dsn:
        dsn = _DEFAULT_GUI_DSN
    if not dsn:
        return

    try:
        import sentry_sdk  # type: ignore[import]
        from sentry_sdk.integrations.logging import LoggingIntegration  # type: ignore[import]
        try:
            # Qt integration is optional – if unavailable, we still proceed
            from sentry_sdk.integrations.qt import QtIntegration  # type: ignore[import]
        except Exception:
            QtIntegration = None  # type: ignore[assignment]
        try:
            # Faulthandler integration can capture hard crashes (e.g. segfaults)
            from sentry_sdk.integrations.faulthandler import (  # type: ignore[import]
                FaulthandlerIntegration,
            )
        except Exception:
            FaulthandlerIntegration = None  # type: ignore[assignment]
    except Exception:
        # Sentry SDK not installed
        return

    logging_integration = LoggingIntegration(
        level=None,        # Do not auto-capture all logs
        event_level=None,  # Only explicit captures / unhandled exceptions
    )

    integrations = [logging_integration]
    try:
        if QtIntegration is not None:
            integrations.append(QtIntegration())
    except Exception:
        # If QtIntegration fails for any reason, continue with logging only
        pass
    try:
        if FaulthandlerIntegration is not None:
            integrations.append(FaulthandlerIntegration())
    except Exception:
        # Hard-crash capture is best-effort; never break startup
        pass

    try:
        sentry_sdk.init(
            dsn=dsn,
            release=release,
            environment=environment,
            integrations=integrations,
            before_send=_before_send,
            attach_stacktrace=True,
            send_default_pii=False,
        )

        # Tag events as originating from the svVascularize GUI
        with sentry_sdk.configure_scope() as scope:
            scope.set_tag("application", "svVascularize")
            scope.set_tag("component", "gui")

        _ENABLED = True

        # Install a simple global excepthook as a last-resort safety net.
        if not _EXCEPTHOOK_INSTALLED:
            _install_global_excepthook()
            _EXCEPTHOOK_INSTALLED = True
    except Exception:
        # Any failure here should not break the application
        _ENABLED = False


def _install_global_excepthook() -> None:
    """Install a global sys.excepthook that reports to Sentry when enabled."""
    try:
        import sentry_sdk  # type: ignore[import]
    except Exception:
        return

    original_hook = sys.excepthook

    def _hook(exc_type, exc_value, exc_traceback):
        # Avoid swallowing KeyboardInterrupt
        if exc_type is KeyboardInterrupt:
            if original_hook:
                original_hook(exc_type, exc_value, exc_traceback)
            return
        try:
            if _ENABLED:
                sentry_sdk.capture_exception(exc_value)
        except Exception:
            pass
        if original_hook:
            original_hook(exc_type, exc_value, exc_traceback)

    sys.excepthook = _hook


def capture_exception(exc: BaseException) -> None:
    """
    Manually capture an exception and send to telemetry backend, if enabled.
    """
    if not _ENABLED:
        return
    try:
        import sentry_sdk  # type: ignore[import]

        sentry_sdk.capture_exception(exc)
        # Flush to ensure the event is sent immediately
        sentry_sdk.flush(timeout=2.0)
    except Exception:
        pass


def capture_message(message: str, level: str = "info", **tags) -> None:
    """
    Manually capture a message (breadcrumb/event) if telemetry is enabled.

    Parameters
    ----------
    message : str
        Message text to record.
    level : str
        Sentry level string, e.g., "info", "warning", "error".
    **tags :
        Optional key/value tags to attach to the event.
    """
    if not _ENABLED:
        return
    try:
        import sentry_sdk  # type: ignore[import]

        with sentry_sdk.push_scope() as scope:
            for key, value in tags.items():
                scope.set_tag(key, value)
            sentry_sdk.capture_message(message, level=level)
        # Flush to ensure the event is sent immediately
        sentry_sdk.flush(timeout=2.0)
    except Exception:
        pass


def telemetry_enabled() -> bool:
    """
    Return True if telemetry/Sentry has been initialized for this process.

    This is primarily intended for GUI debug helpers so they can inform the
    user when a telemetry test event could not be sent because telemetry is
    disabled or the Sentry SDK is unavailable.
    """
    return _ENABLED



import numpy as np
import pytest


def _fake_run_mmg(tool, args, stdout=None, stderr=None, cwd=None):
    import meshio
    from pathlib import Path

    assert tool == "mmgs"
    assert cwd is not None
    cwd_path = Path(cwd)

    in_mesh = cwd_path / str(args[0])
    text = in_mesh.read_text(errors="ignore")
    # Regression check: MMGS input must contain triangles (e.g., pv.Cube() starts as quads).
    assert "Triangles" in text

    points = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]], dtype=float)
    triangles = np.array([[0, 1, 2]], dtype=int)
    mesh = meshio.Mesh(points, cells=[("triangle", triangles)])
    meshio.write(str(cwd_path / "tmp.o.mesh"), mesh, file_format="medit")
    return cwd_path / "mmgs_O3"


def test_remesh_surface_triangulates_quads(monkeypatch, tmp_path):
    import pyvista as pv
    from svv.utils.remeshing import remesh as remesh_mod

    monkeypatch.setattr(remesh_mod, "run_mmg", _fake_run_mmg)
    monkeypatch.chdir(tmp_path)

    out = remesh_mod.remesh_surface(pv.Cube(), autofix=False, verbosity=0)
    assert out.is_all_triangles


def test_remesh_surface_required_triangles_requires_tri_mesh():
    import pyvista as pv
    from svv.utils.remeshing import remesh as remesh_mod

    with pytest.raises(ValueError, match="requires the input surface to be all triangles"):
        remesh_mod.remesh_surface(pv.Cube(), required_triangles=[1], autofix=False, verbosity=0)


def test_remesh_surface_deletes_in_sol_on_success(monkeypatch, tmp_path):
    import pyvista as pv
    from svv.utils.remeshing import remesh as remesh_mod

    monkeypatch.setattr(remesh_mod, "run_mmg", _fake_run_mmg)
    monkeypatch.chdir(tmp_path)

    (tmp_path / "in.sol").write_text("dummy sizing function\n")
    remesh_mod.remesh_surface(pv.Cube(), autofix=False, verbosity=0)
    assert not (tmp_path / "in.sol").exists()




import numpy as np
import pyvista as pv

from svv.domain.domain import Domain


def _make_tet_mesh(points: np.ndarray, tets: np.ndarray) -> pv.UnstructuredGrid:
    points = np.asarray(points, dtype=float)
    tets = np.asarray(tets, dtype=np.int64)
    n_cells = int(tets.shape[0])
    cells = np.hstack([np.full((n_cells, 1), 4, dtype=np.int64), tets]).ravel()
    celltypes = np.full(n_cells, pv.CellType.TETRA, dtype=np.uint8)
    return pv.UnstructuredGrid(cells, celltypes, points)


def _dummy_domain_with_mesh(mesh: pv.UnstructuredGrid) -> Domain:
    dom = Domain(np.zeros((1, 3), dtype=float))
    dom.mesh = mesh
    # Provide a cheap implicit evaluator so voronoi sampling can filter by implicit_range.
    dom.evaluate_fast = lambda pts, **_: -0.5 * np.ones((np.asarray(pts).shape[0], 1), dtype=float)
    return dom


def test_voronoi_sampling_is_deterministic_and_uses_cell_centers():
    # Five tetrahedra with distinct cell centers.
    points = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [3, 0, 0],
            [4, 0, 0],
            [3, 1, 0],
            [3, 0, 1],
            [0, 3, 0],
            [0, 4, 0],
            [1, 3, 0],
            [0, 3, 1],
            [0, 0, 3],
            [1, 0, 3],
            [0, 1, 3],
            [0, 0, 4],
            [2, 2, 2],
            [3, 2, 2],
            [2, 3, 2],
            [2, 2, 3],
        ],
        dtype=float,
    )
    tets = np.array(
        [
            [0, 1, 2, 3],      # near origin
            [4, 5, 6, 7],      # +x
            [8, 9, 10, 11],    # +y
            [12, 13, 14, 15],  # +z
            [16, 17, 18, 19],  # diagonal
        ],
        dtype=np.int64,
    )
    mesh = _make_tet_mesh(points, tets)
    dom = _dummy_domain_with_mesh(mesh)

    seeds = np.array([[0.0, 0.0, 0.0]])

    pts1, cells1 = dom.get_interior_points(3, tree=seeds, threshold=0.0, method="voronoi", implicit_range=(-1.0, 0.0))
    pts2, cells2 = dom.get_interior_points(3, tree=seeds, threshold=0.0, method="voronoi", implicit_range=(-1.0, 0.0))

    assert np.allclose(pts1, pts2, equal_nan=True)
    assert np.array_equal(cells1, cells2)

    centers = mesh.cell_centers().points
    for p, cid in zip(pts1, cells1):
        assert cid >= 0
        assert np.allclose(p, centers[cid])

    # Farthest-first order should pick the diagonal cell, then +x, then +y given the stable tie-breaks.
    assert cells1.tolist() == [4, 1, 2]


def test_voronoi_sampling_respects_volume_threshold():
    points = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [3, 0, 0],
            [4, 0, 0],
            [3, 1, 0],
            [3, 0, 1],
            [2, 2, 2],
            [3, 2, 2],
            [2, 3, 2],
            [2, 2, 3],
        ],
        dtype=float,
    )
    tets = np.array(
        [
            [0, 1, 2, 3],     # near origin
            [4, 5, 6, 7],     # +x (dist ~3.268)
            [8, 9, 10, 11],   # diagonal (dist ~3.897)
        ],
        dtype=np.int64,
    )
    mesh = _make_tet_mesh(points, tets)
    dom = _dummy_domain_with_mesh(mesh)

    seeds = np.array([[0.0, 0.0, 0.0]])

    # Exclude the diagonal cell by limiting the allowed distance to the existing tree.
    pts, cells = dom.get_interior_points(
        2,
        tree=seeds,
        threshold=0.0,
        volume_threshold=3.3,
        method="voronoi",
        implicit_range=(-1.0, 0.0),
    )

    assert np.array_equal(cells, np.array([1, 0], dtype=np.int64))
    centers = mesh.cell_centers().points
    assert np.allclose(pts[0], centers[1])
    assert np.allclose(pts[1], centers[0])



"""
Example usage of the svVascularize GUI.

This script demonstrates how to launch the GUI with and without a domain.
"""
import sys
from PySide6.QtWidgets import QApplication
from svv.visualize.gui import VascularizeGUI


def example_empty():
    """Launch GUI without a domain (domain can be loaded from menu)."""
    app = QApplication(sys.argv)
    gui = VascularizeGUI()
    gui.show()
    try:
        sys.exit(app.exec())
    except AttributeError:
        sys.exit(app.exec_())


def example_with_domain():
    """Launch GUI with a pre-loaded domain."""
    import pyvista as pv
    from svv.domain.domain import Domain

    # Create a simple cube domain
    cube_mesh = pv.Cube()
    domain = Domain(cube_mesh)
    domain.create()
    domain.solve()
    domain.build()

    # Launch GUI with domain
    app = QApplication(sys.argv)
    gui = VascularizeGUI(domain=domain)
    gui.show()
    try:
        sys.exit(app.exec())
    except AttributeError:
        sys.exit(app.exec_())


def example_with_sphere():
    """Launch GUI with a sphere domain."""
    import pyvista as pv
    from svv.domain.domain import Domain

    # Create a sphere domain
    sphere = pv.Sphere(radius=5.0, theta_resolution=30, phi_resolution=30)
    domain = Domain(sphere)
    domain.create()
    domain.solve()
    domain.build()

    # Launch GUI with domain
    app = QApplication(sys.argv)
    gui = VascularizeGUI(domain=domain)
    gui.show()
    try:
        sys.exit(app.exec())
    except AttributeError:
        sys.exit(app.exec_())


def example_load_from_file(dmn_file):
    """Launch GUI with a domain loaded from a .dmn file."""
    from svv.domain.domain import Domain

    # Load domain from file
    domain = Domain.load(dmn_file)

    # Launch GUI with domain
    app = QApplication(sys.argv)
    gui = VascularizeGUI(domain=domain)
    gui.show()
    try:
        sys.exit(app.exec())
    except AttributeError:
        sys.exit(app.exec_())


if __name__ == "__main__":
    # Choose which example to run:

    # 1. Empty GUI (load domain from File menu)
    # example_empty()

    # 2. GUI with a cube domain
    example_with_domain()

    # 3. GUI with a sphere domain
    # example_with_sphere()

    # 4. GUI with domain loaded from file
    # example_load_from_file("path/to/your/domain.dmn")



"""
Platform/architecture-aware MMG executable selection.

This module intentionally avoids importing heavy dependencies (PyVista, VTK, etc.)
so it can be used in lightweight CI checks and during wheel build validation.
"""

from __future__ import annotations

import errno
import hashlib
import os
import platform
import shutil
import stat
import subprocess
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Sequence, Tuple


MMG_TOOL_TO_EXE = {
    "mmg2d": "mmg2d_O3",
    "mmg3d": "mmg3d_O3",
    "mmgs": "mmgs_O3",
}


def _norm_os_dir(system: Optional[str] = None) -> str:
    sysname = (system or platform.system()).strip()
    if sysname == "Linux":
        return "Linux"
    if sysname == "Windows":
        return "Windows"
    if sysname == "Darwin":
        return "Mac"
    raise RuntimeError(f"Unsupported OS: {sysname}")


def _norm_arch(machine: Optional[str] = None) -> str:
    m = (machine or platform.machine()).strip().lower()
    if m in {"x86_64", "amd64"}:
        return "x86_64"
    if m in {"aarch64", "arm64"}:
        return "aarch64"
    return m or "unknown"


def _mmg_exe_filename(tool: str) -> str:
    if tool not in MMG_TOOL_TO_EXE:
        raise ValueError(f"Unknown MMG tool: {tool!r} (expected one of {sorted(MMG_TOOL_TO_EXE)})")
    base = MMG_TOOL_TO_EXE[tool]
    if os.name == "nt":
        return base + ".exe"
    return base


def _ensure_executable(path: Path) -> None:
    if os.name == "nt":
        return
    try:
        mode = path.stat().st_mode
        if mode & stat.S_IXUSR:
            return
        path.chmod(mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    except Exception:
        # Best-effort; actual execution will raise if this fails.
        pass


def _is_exec_format_error(e: BaseException) -> bool:
    if isinstance(e, OSError):
        if getattr(e, "errno", None) == errno.ENOEXEC:
            return True
        # Windows: ERROR_BAD_EXE_FORMAT (193)
        if getattr(e, "winerror", None) == 193:
            return True
    return False


def _uid_suffix() -> str:
    try:
        return str(os.getuid())
    except Exception:
        return "unknown"


def _user_cache_dir() -> Optional[Path]:
    try:
        if os.name == "nt":
            base = os.environ.get("LOCALAPPDATA") or os.environ.get("APPDATA")
            if not base:
                return None
            return Path(base) / "svv" / "mmg"
        base = os.environ.get("XDG_CACHE_HOME")
        if base:
            return Path(base) / "svv" / "mmg"
        return Path.home() / ".cache" / "svv" / "mmg"
    except Exception:
        return None


def _mmg_run_roots() -> Tuple[Path, ...]:
    roots: List[Path] = []
    env_root = os.environ.get("SVV_MMG_RUN_DIR")
    if env_root:
        roots.append(Path(env_root).expanduser())
    try:
        roots.append(Path(tempfile.gettempdir()))
    except Exception:
        pass
    cache = _user_cache_dir()
    if cache is not None:
        roots.append(cache)
    # De-duplicate while preserving order
    uniq: List[Path] = []
    seen = set()
    for r in roots:
        key = str(r)
        if key in seen:
            continue
        seen.add(key)
        uniq.append(r)
    return tuple(uniq)


def _sha256_short(path: Path) -> Optional[str]:
    try:
        h = hashlib.sha256()
        with path.open("rb") as f:
            for chunk in iter(lambda: f.read(1024 * 1024), b""):
                h.update(chunk)
        return h.hexdigest()[:12]
    except Exception:
        return None


@dataclass(frozen=True)
class MMGSelection:
    tool: str
    os_dir: str
    arch: str
    candidates: Tuple[Path, ...]


def get_mmg_candidates(
    tool: str,
    *,
    os_dir: Optional[str] = None,
    arch: Optional[str] = None,
) -> MMGSelection:
    """
    Return an ordered list of candidate executable paths for the requested MMG tool.

    Search order:
    1) SVV_MMG_PATH (explicit executable)
    2) SVV_MMG_DIR/<exe>
    3) svv/bin/<exe> (locally built override)
    4) bundled: svv/utils/remeshing/<OS>/<arch>/<exe>
    5) bundled fallbacks (mac): universal2 -> arch -> legacy
    6) legacy flat paths: svv/utils/remeshing/<OS>/<exe>
    """
    os_dir = os_dir or _norm_os_dir()
    arch_override = os.environ.get("SVV_MMG_ARCH")
    arch = (arch_override or arch or "").strip()
    if not arch:
        arch = _norm_arch()
    else:
        a = arch.lower()
        if a in {"amd64", "x86_64"}:
            arch = "x86_64"
        elif a in {"arm64", "aarch64"}:
            arch = "aarch64"
        elif a in {"universal2"}:
            arch = "universal2"

    exe_name = _mmg_exe_filename(tool)

    # Base package dir: .../svv/utils/remeshing
    remesh_dir = Path(__file__).resolve().parent

    candidates: List[Path] = []

    mmg_path = os.environ.get("SVV_MMG_PATH")
    if mmg_path:
        candidates.append(Path(mmg_path).expanduser())

    mmg_dir = os.environ.get("SVV_MMG_DIR")
    if mmg_dir:
        candidates.append(Path(mmg_dir).expanduser() / exe_name)

    # svv/bin override if present
    try:
        import svv  # imported lazily to keep this module lightweight

        svv_root = Path(svv.__file__).resolve().parent
        candidates.append(svv_root / "bin" / exe_name)
    except Exception:
        pass

    # Bundled per-platform layout.
    os_base = remesh_dir / os_dir
    if os_dir == "Mac":
        # Prefer universal2 if present, then machine arch fallbacks.
        candidates.append(os_base / "universal2" / exe_name)
        # Fallback: some trees may use x86_64/arm64 or x86_64/aarch64 naming.
        candidates.append(os_base / "x86_64" / exe_name)
        candidates.append(os_base / "arm64" / exe_name)
        candidates.append(os_base / "aarch64" / exe_name)
    else:
        candidates.append(os_base / arch / exe_name)

    # Legacy flat layout (pre-arch split)
    candidates.append(os_base / exe_name)

    # De-duplicate while preserving order
    uniq: List[Path] = []
    seen = set()
    for p in candidates:
        key = str(p)
        if key in seen:
            continue
        seen.add(key)
        uniq.append(p)

    return MMGSelection(tool=tool, os_dir=os_dir, arch=arch, candidates=tuple(uniq))


def get_mmg_exe(tool: str) -> Path:
    sel = get_mmg_candidates(tool)
    for p in sel.candidates:
        if p.is_file():
            return p
    tried = "\n  - ".join(str(p) for p in sel.candidates)
    raise FileNotFoundError(
        f"MMG executable not found for tool={tool!r} (os={sel.os_dir}, arch={sel.arch}). Tried:\n  - {tried}"
    )


def run_mmg(
    tool: str,
    args: Sequence[str],
    *,
    stdout=None,
    stderr=None,
    cwd=None,
) -> Path:
    """
    Execute the requested MMG tool, trying platform/arch candidates as needed.

    Returns the Path of the executable that successfully ran.
    """
    sel = get_mmg_candidates(tool)
    last_exec_error: Optional[BaseException] = None

    for exe in sel.candidates:
        if not exe.is_file():
            continue
        try:
            _ensure_executable(exe)
            subprocess.check_call([str(exe), *map(str, args)], stdout=stdout, stderr=stderr, cwd=cwd)
            return exe
        except subprocess.CalledProcessError:
            # Tool ran but failed; this is not an architecture-selection issue.
            raise
        except PermissionError as e:
            # If we can't exec in-place (missing +x, read-only site-packages, or a noexec mount),
            # try copying the executable into a user-writable run directory and executing from there.
            last_exec_error = e
            digest = _sha256_short(exe)
            for root in _mmg_run_roots():
                try:
                    base = root / f"svv-mmg-{_uid_suffix()}" / sel.os_dir / sel.arch
                    if digest:
                        base = base / digest
                    base.mkdir(parents=True, exist_ok=True)
                    if os.name != "nt":
                        try:
                            base.chmod(0o700)
                        except Exception:
                            pass
                    cached = base / exe.name
                    try:
                        if (not cached.is_file()) or (cached.stat().st_size != exe.stat().st_size):
                            shutil.copy2(exe, cached)
                    except Exception:
                        # If copy/overwrite fails (permissions, races), try running whatever exists.
                        pass
                    _ensure_executable(cached)
                    subprocess.check_call([str(cached), *map(str, args)], stdout=stdout, stderr=stderr, cwd=cwd)
                    return cached
                except subprocess.CalledProcessError:
                    raise
                except BaseException as copy_exc:
                    if isinstance(copy_exc, (FileNotFoundError, PermissionError)) or _is_exec_format_error(copy_exc):
                        last_exec_error = copy_exc
                        continue
                    # Non-exec-related error: surface it.
                    raise
            continue
        except BaseException as e:
            # Exec-format mismatch: try another candidate.
            if _is_exec_format_error(e):
                last_exec_error = e
                continue
            raise

    tried = "\n  - ".join(str(p) for p in sel.candidates)
    if last_exec_error is not None:
        raise RuntimeError(
            f"No compatible MMG executable could be executed for tool={tool!r} "
            f"(os={sel.os_dir}, arch={sel.arch}). Tried:\n  - {tried}\n"
            f"Last error: {last_exec_error}"
        ) from last_exec_error
    raise FileNotFoundError(
        f"MMG executable not found for tool={tool!r} (os={sel.os_dir}, arch={sel.arch}). Tried:\n  - {tried}"
    )



import numpy as np

def get_angles(vector1: np.ndarray, vector2: np.ndarray) -> np.ndarray:
    v1 = np.asarray(vector1, dtype=float)
    v2 = np.asarray(vector2, dtype=float)
    dot = np.einsum('ij,ij->i', v1, v2)
    n1 = np.linalg.norm(v1, axis=1)
    n2 = np.linalg.norm(v2, axis=1)
    denom = n1 * n2

    # Match Cython behavior: if denom < 1e-15, return angle = 0.0
    # Otherwise compute arccos of clamped ratio
    angles = np.zeros(len(v1), dtype=np.float64)
    valid = denom >= 1e-15

    if np.any(valid):
        with np.errstate(invalid='ignore', divide='ignore'):
            ratio = dot[valid] / denom[valid]
        ratio = np.clip(ratio, -1.0, 1.0)
        angles[valid] = np.degrees(np.arccos(ratio))

    return angles




from svv.forest.connect.tree_connection import TreeConnection
import numpy as np
from scipy.spatial import cKDTree


class ForestConnection:
    def __init__(self, forest, **kwargs):
        self.forest = forest
        self.tree_connections = []
        self.networks = []
        self._static_collision_cache = None

    @staticmethod
    def _build_static_collision_cache(forest):
        """
        Build a cached array of all tree vessel line segments in the forest.

        The result is used by TreeConnection to avoid rebuilding and copying
        the full collision segment list for every individual connection.
        """
        tree_ranges = [
            [slice(0, 0) for _ in range(forest.n_trees_per_network[i])]
            for i in range(forest.n_networks)
        ]
        segments_by_tree = []
        offset = 0

        for net_idx in range(forest.n_networks):
            for tree_idx in range(forest.n_trees_per_network[net_idx]):
                data = np.asarray(forest.networks[net_idx][tree_idx].data)
                if data.ndim != 2 or data.shape[0] == 0:
                    tree_ranges[net_idx][tree_idx] = slice(offset, offset)
                    continue
                n_rows = int(data.shape[0])
                segs = np.zeros((n_rows, 7), dtype=float)
                segs[:, 0:3] = data[:, 0:3]
                segs[:, 3:6] = data[:, 3:6]
                segs[:, 6] = data[:, 21]
                segments_by_tree.append(segs)
                tree_ranges[net_idx][tree_idx] = slice(offset, offset + n_rows)
                offset += n_rows

        if offset == 0:
            segments = np.zeros((0, 7), dtype=float)
            endpoints = np.zeros((0, 3), dtype=float)
            endpoint_tree = None
            endpoint_to_segment = np.zeros((0,), dtype=int)
        else:
            segments = np.vstack(segments_by_tree)
            endpoints = np.vstack([segments[:, 0:3], segments[:, 3:6]])
            endpoint_to_segment = np.concatenate([np.arange(offset, dtype=int), np.arange(offset, dtype=int)])
            endpoint_tree = cKDTree(endpoints)

        return {
            "segments": segments,
            "tree_ranges": tree_ranges,
            "endpoint_tree": endpoint_tree,
            "endpoint_to_segment": endpoint_to_segment,
        }

    def solve(self, *args, num_vessels=20, attempts=5, **kwargs):
        self.tree_connections = []
        if self._static_collision_cache is None:
            self._static_collision_cache = self._build_static_collision_cache(self.forest)
        for i in range(self.forest.n_networks):
            tree_connections = TreeConnection(self.forest, i, collision_cache=self._static_collision_cache, **kwargs)
            if len(self.tree_connections) > 0:
                other_vessels = []
                for j in range(len(self.tree_connections)):
                    other_vessels.extend(self.tree_connections[j].vessels)
                tree_connections.other_vessels = other_vessels
            tree_connections.solve(*args, num_vessels=num_vessels, attempts=attempts)
            self.tree_connections.append(tree_connections)

    def export_solid(self, cap_resolution=40, extrude_roots=False):
        network_solids = []
        network_lines = []
        network_tubes = []
        for i in range(len(self.tree_connections)):
            network_solid, network_line, network_tube = self.tree_connections[i].export_solid(cap_resolution=cap_resolution,
                                                                                         extrude_roots=extrude_roots)
            network_solids.extend(network_solid)
            network_lines.extend(network_line)
            network_tubes.extend(network_tube)
        return network_solids, network_lines, network_tubes

    def show(self, **kwargs):
        pass



"""
Parameter panel for configuring Tree and Forest generation.
"""
from PySide6.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QHBoxLayout, QGroupBox,
    QPushButton, QLabel, QSpinBox, QDoubleSpinBox,
    QComboBox, QCheckBox, QScrollArea, QFormLayout,
    QMessageBox, QProgressDialog, QSizePolicy,
    QToolButton, QToolTip
)
import numpy as np
import threading
from queue import Queue
from PySide6.QtCore import Signal, Qt, QTimer, QSignalBlocker, QSettings, QSize
from concurrent.futures import ThreadPoolExecutor
from svv.visualize.gui.theme import CADTheme, CADIcons
from svv.tree.data.data import TreeParameters
from svv.tree.data.units import UnitSystem, _LENGTH_UNITS, _MASS_UNITS, _TIME_UNITS
import traceback
from svv.telemetry import capture_exception, capture_message

# Descriptions shown in info-icon popups beside each parameter label.
_PARAM_DESCRIPTIONS = {
    'unit_system': (
        "Base units (Length, Mass, Time) for all dimensional parameters.\n"
        "Changing units will automatically convert existing values\n"
        "to maintain physical consistency."
    ),
    'n_vessels': (
        "Total number of terminal vessel segments to generate.\n"
        "Each vessel is a branch endpoint in the vascular tree.\n"
        "More vessels produce denser, more detailed networks."
    ),
    'physical_clearance': (
        "Minimum distance maintained between vessel walls during\n"
        "generation. Set to 0 to allow vessels to touch or overlap.\n"
        "Larger values create more spaced-out networks."
    ),
    'convexity_tolerance': (
        "Controls how strictly the domain convexity is checked when\n"
        "placing new vessels. Smaller values enforce stricter convexity,\n"
        "while larger values allow more flexibility near concave regions."
    ),
    'kinematic_viscosity': (
        "Dynamic viscosity divided by fluid density (\u03bd = \u03bc/\u03c1).\n"
        "Controls the viscous resistance of blood flow through vessels.\n"
        "Affects vessel radius optimization via Poiseuille's law."
    ),
    'fluid_density': (
        "Mass density of the fluid (blood). Used together with viscosity\n"
        "to compute flow resistance and pressure drops through the\n"
        "vascular network."
    ),
    'murray_exponent': (
        "Exponent in Murray's law governing the relationship between\n"
        "parent and child vessel radii at bifurcations. A value of 3\n"
        "corresponds to optimal laminar flow. Values > 3 favor thinner\n"
        "child branches."
    ),
    'radius_exponent': (
        "Exponent controlling how vessel radius scales with tree level.\n"
        "Affects the tapering profile of the vascular tree."
    ),
    'length_exponent': (
        "Exponent controlling how vessel segment lengths scale relative\n"
        "to their radii. Affects the overall shape and density of the\n"
        "tree structure."
    ),
    'terminal_pressure': (
        "Blood pressure at the terminal (outlet) ends of the vascular\n"
        "tree. Together with root pressure, determines the pressure\n"
        "gradient driving flow through the network."
    ),
    'root_pressure': (
        "Blood pressure at the root (inlet) of the vascular tree.\n"
        "The pressure difference between root and terminal pressures\n"
        "drives flow through the entire network."
    ),
    'terminal_flow': (
        "Volumetric flow rate at each terminal vessel segment.\n"
        "Determines how much blood exits at each branch endpoint.\n"
        "Total network flow equals terminal_flow \u00d7 number of terminals."
    ),
    'root_flow': (
        "Total volumetric flow rate entering at the root vessel.\n"
        "When set to Auto, computed as the sum of all terminal flows.\n"
        "Set manually to override the automatic calculation."
    ),
    'max_nonconvex_count': (
        "Maximum number of candidate positions sampled in non-convex\n"
        "regions before the algorithm moves on. Higher values allow\n"
        "better vessel placement in complex geometries but increase\n"
        "computation time."
    ),
    'n_networks': (
        "Number of independent vascular networks in the forest.\n"
        "Each network is a separate connected vascular system\n"
        "(e.g., arterial and venous networks)."
    ),
    'trees_per_network': (
        "Number of trees in each network. Trees within a network are\n"
        "later connected to form a complete vascular circuit. Each tree\n"
        "grows from its own start point."
    ),
    'vessels_per_network': (
        "Number of vessel-growth steps to run for each network.\n"
        "This controls how many vessels are added to every tree in the\n"
        "network during forest generation. Networks can stop growing\n"
        "earlier than others by setting a smaller value here."
    ),
    'connection_curve': (
        "Curve type used to connect trees within each network.\n"
        "\u2022 Bezier: Smooth polynomial curves\n"
        "\u2022 Catmull-Rom: Interpolating splines through control points\n"
        "\u2022 NURBS: Non-Uniform Rational B-Splines for maximum flexibility"
    ),
    'competition': (
        "When enabled, trees compete for territory during growth.\n"
        "Trees encountering another tree's territory stop expanding\n"
        "in that direction, creating natural boundaries between trees."
    ),
    'decay_probability': (
        "Controls how aggressively trees yield territory during\n"
        "competition. Higher values (closer to 1.0) make trees decay\n"
        "faster at boundaries, creating cleaner separations. Lower\n"
        "values allow more territorial overlap."
    ),
    'random_seed': (
        "Seed for the random number generator. Set to a specific value\n"
        "for reproducible results. 'Random' uses a different seed\n"
        "each run."
    ),
    'preallocation_size': (
        "Number of rows preallocated in memory for each tree's internal\n"
        "data arrays. Higher values support more vessels but consume\n"
        "more memory. Lower values are more memory-efficient but may\n"
        "limit the maximum tree size."
    ),
}


class ParameterPanel(QWidget):
    """
    Widget for configuring Tree and Forest generation parameters.
    """

    # Signals
    parameters_changed = Signal()
    length_unit_changed = Signal(str)  # Emitted when length unit changes (e.g., "cm", "mm", "m")

    def __init__(self, parent=None):
        """
        Initialize the parameter panel.

        Parameters
        ----------
        parent : QWidget, optional
            Parent widget (should be VascularizeGUI)
        """
        super().__init__(parent)
        self.parent_gui = parent

        self._connect_future = None
        self._connect_progress = None
        self._executor = ThreadPoolExecutor(max_workers=1)
        self.tree_param_overrides = {}
        base_tree_params = TreeParameters()
        self.tree_unit_system = base_tree_params.unit_system
        self.tree_param_units = self._build_tree_param_units()
        self.tree_param_unit_labels = {}
        self.tree_base_params = self._default_tree_params(base_tree_params)
        self._generation_future = None
        self._generation_progress = None
        self._generation_progress_queue = None
        self._generation_cancel_event = None

        self._init_ui()

    def _record_telemetry(self, exc=None, message=None, level="error", traceback_str=None, **tags):
        """Send GUI warnings/errors to telemetry without impacting the UI.

        Parameters
        ----------
        exc : Exception, optional
            Exception to capture. If provided, captures as an exception event.
        message : str, optional
            Message to capture. Used when exc is None.
        level : str
            Sentry level ("error", "warning", "info").
        traceback_str : str, optional
            Full traceback string to include as extra context.
        **tags
            Additional tags to attach to the event.
        """
        try:
            if exc is not None:
                try:
                    import sentry_sdk  # type: ignore[import]

                    with sentry_sdk.push_scope() as scope:
                        for key, value in tags.items():
                            scope.set_tag(key, value)
                        if traceback_str:
                            scope.set_extra("full_traceback", traceback_str)
                        sentry_sdk.capture_exception(exc)
                        # Flush to ensure the event is sent before the popup blocks
                        sentry_sdk.flush(timeout=2.0)
                except Exception:
                    capture_exception(exc)
                return
            if message:
                capture_message(message, level=level, **tags)
        except Exception:
            pass

    def shutdown(self):
        """
        Cancel any in-flight generation/connect tasks and shut down the worker
        executor so threads and memory are released when the application closes.
        """
        # Signal cancellation to any running tasks
        if getattr(self, "_generation_cancel_event", None) is not None:
            try:
                self._generation_cancel_event.set()
            except Exception:
                pass
        if getattr(self, "_connect_future", None) is not None:
            self._connect_future = None
        # Shut down executor threads
        try:
            self._executor.shutdown(wait=False, cancel_futures=True)
        except TypeError:
            self._executor.shutdown(wait=False)
        except Exception:
            pass

    def _init_ui(self):
        """Initialize the user interface."""
        layout = QVBoxLayout()
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setSpacing(8)
        self.setLayout(layout)

        # Create scroll area for parameters
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_content = QWidget()
        scroll_layout = QVBoxLayout(scroll_content)
        scroll_layout.setSpacing(12)
        # Add right margin to account for scrollbar width (~16-20px)
        scroll_layout.setContentsMargins(0, 0, 20, 0)

        # Mode selection
        mode_group = QGroupBox("Generation Mode")
        mode_layout = QVBoxLayout()
        mode_group.setLayout(mode_layout)

        self.mode_combo = QComboBox()
        self.mode_combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.mode_combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.mode_combo.addItem("Tree (single)")
        self.mode_combo.addItem("Forest (multiple)")
        self.mode_combo.setMaxVisibleItems(100)
        self.mode_combo.currentIndexChanged.connect(self._on_mode_changed)
        self.mode_combo.setToolTip("Select whether to generate a single tree or a forest with multiple trees")
        mode_layout.addWidget(self.mode_combo)

        scroll_layout.addWidget(mode_group)

        # Tree parameters
        tree_params_group = QGroupBox("Tree Parameters")
        tree_form = QFormLayout()
        tree_form.setSpacing(10)
        tree_form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        tree_params_group.setLayout(tree_form)

        # Unit system selection (dropdowns for base units)
        unit_row = QWidget()
        unit_layout = QHBoxLayout(unit_row)
        unit_layout.setContentsMargins(0, 0, 0, 0)
        unit_layout.setSpacing(6)

        self.length_unit_combo = QComboBox()
        self.length_unit_combo.addItems(sorted(_LENGTH_UNITS.keys()))
        self.length_unit_combo.setCurrentText(self.tree_unit_system.base.length.symbol)
        self.length_unit_combo.setToolTip("Base length unit used for TreeParameters values")
        self.length_unit_combo.currentTextChanged.connect(self._on_unit_system_changed)
        unit_layout.addWidget(QLabel("L:"))
        unit_layout.addWidget(self.length_unit_combo)

        self.mass_unit_combo = QComboBox()
        self.mass_unit_combo.addItems(sorted(_MASS_UNITS.keys()))
        self.mass_unit_combo.setCurrentText(self.tree_unit_system.base.mass.symbol)
        self.mass_unit_combo.setToolTip("Base mass unit used for TreeParameters values")
        self.mass_unit_combo.currentTextChanged.connect(self._on_unit_system_changed)
        unit_layout.addWidget(QLabel("M:"))
        unit_layout.addWidget(self.mass_unit_combo)

        self.time_unit_combo = QComboBox()
        self.time_unit_combo.addItems(sorted(_TIME_UNITS.keys()))
        self.time_unit_combo.setCurrentText(self.tree_unit_system.base.time.symbol)
        self.time_unit_combo.setToolTip("Base time unit used for TreeParameters values")
        self.time_unit_combo.currentTextChanged.connect(self._on_unit_system_changed)
        unit_layout.addWidget(QLabel("T:"))
        unit_layout.addWidget(self.time_unit_combo)

        unit_layout.addStretch()
        tree_form.addRow(self._make_label_with_info("Unit System:", _PARAM_DESCRIPTIONS['unit_system']), unit_row)

        # Number of vessels
        self.n_vessels_spin = QSpinBox()
        self.n_vessels_spin.setRange(1, 100000)
        self.n_vessels_spin.setValue(100)
        self.n_vessels_spin.setToolTip("Total number of vessel segments to generate")
        tree_form.addRow(
            self._make_label_with_info("Number of Vessels:", _PARAM_DESCRIPTIONS['n_vessels']),
            self.n_vessels_spin
        )

        # Physical clearance
        self.physical_clearance_spin = QDoubleSpinBox()
        self.physical_clearance_spin.setRange(0.0, 10.0)
        self.physical_clearance_spin.setSingleStep(0.01)
        self.physical_clearance_spin.setDecimals(4)
        self.physical_clearance_spin.setValue(0.0)
        self.physical_clearance_spin.setToolTip("Minimum distance between vessel walls (0 = allow touching)")
        tree_form.addRow(
            self._make_label_with_info("Physical Clearance:", _PARAM_DESCRIPTIONS['physical_clearance']),
            self.physical_clearance_spin
        )

        # Convexity tolerance
        self.convexity_tolerance_spin = QDoubleSpinBox()
        self.convexity_tolerance_spin.setRange(0.0, 1.0)
        self.convexity_tolerance_spin.setSingleStep(0.01)
        self.convexity_tolerance_spin.setDecimals(3)
        self.convexity_tolerance_spin.setValue(0.01)
        self.convexity_tolerance_spin.setToolTip("Tolerance for domain convexity checking (smaller = stricter)")
        tree_form.addRow(
            self._make_label_with_info("Convexity Tolerance:", _PARAM_DESCRIPTIONS['convexity_tolerance']),
            self.convexity_tolerance_spin
        )

        # Per-tree override selector (shown in forest mode)
        self.tree_override_widget = QWidget()
        override_layout = QHBoxLayout()
        override_layout.setContentsMargins(0, 0, 0, 0)
        override_layout.setSpacing(6)
        self.tree_override_widget.setLayout(override_layout)
        override_layout.addWidget(QLabel("Edit Tree:"))
        self.override_network_combo = QComboBox()
        self.override_network_combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.override_network_combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.override_network_combo.setMaxVisibleItems(100)
        self.override_tree_combo = QComboBox()
        self.override_tree_combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.override_tree_combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.override_tree_combo.setMaxVisibleItems(100)
        self.override_network_combo.currentIndexChanged.connect(self._on_tree_override_changed)
        self.override_tree_combo.currentIndexChanged.connect(self._on_tree_override_changed)
        override_layout.addWidget(QLabel("Network"))
        override_layout.addWidget(self.override_network_combo)
        override_layout.addWidget(QLabel("Tree"))
        override_layout.addWidget(self.override_tree_combo)
        override_layout.addStretch()
        tree_form.addRow(self.tree_override_widget)
        self.tree_override_widget.hide()

        # TreeParameters fields
        self.tree_param_widgets = {}
        self._add_tree_param_spin(tree_form, 'kinematic_viscosity', "Kinematic Viscosity", 0.0, 10.0, 5, 0.001, self.tree_param_units.get('kinematic_viscosity'), _PARAM_DESCRIPTIONS.get('kinematic_viscosity'))
        self._add_tree_param_spin(tree_form, 'fluid_density', "Fluid Density", 0.0, 20.0, 4, 0.01, self.tree_param_units.get('fluid_density'), _PARAM_DESCRIPTIONS.get('fluid_density'))
        self._add_tree_param_spin(tree_form, 'murray_exponent', "Murray Exponent", 0.5, 10.0, 3, 0.1, self.tree_param_units.get('murray_exponent'), _PARAM_DESCRIPTIONS.get('murray_exponent'))
        self._add_tree_param_spin(tree_form, 'radius_exponent', "Radius Exponent", 0.5, 10.0, 3, 0.1, self.tree_param_units.get('radius_exponent'), _PARAM_DESCRIPTIONS.get('radius_exponent'))
        self._add_tree_param_spin(tree_form, 'length_exponent', "Length Exponent", 0.1, 10.0, 3, 0.1, self.tree_param_units.get('length_exponent'), _PARAM_DESCRIPTIONS.get('length_exponent'))
        self._add_tree_param_spin(tree_form, 'terminal_pressure', "Terminal Pressure", 0.0, 1_000_000.0, 3, 10.0, self.tree_param_units.get('terminal_pressure'), _PARAM_DESCRIPTIONS.get('terminal_pressure'))
        self._add_tree_param_spin(tree_form, 'root_pressure', "Root Pressure", 0.0, 1_000_000.0, 3, 10.0, self.tree_param_units.get('root_pressure'), _PARAM_DESCRIPTIONS.get('root_pressure'))
        self._add_tree_param_spin(tree_form, 'terminal_flow', "Terminal Flow", 0.0, 100.0, 6, 0.0001, self.tree_param_units.get('terminal_flow'), _PARAM_DESCRIPTIONS.get('terminal_flow'))

        root_flow_layout = QHBoxLayout()
        self.root_flow_spin = QDoubleSpinBox()
        self.root_flow_spin.setRange(0.0, 100.0)
        self.root_flow_spin.setDecimals(6)
        self.root_flow_spin.setSingleStep(0.0001)
        self.root_flow_spin.valueChanged.connect(lambda v: self._on_tree_param_changed('root_flow', v))
        self.root_flow_auto_cb = QCheckBox("Auto")
        self.root_flow_auto_cb.setChecked(True)
        self.root_flow_auto_cb.stateChanged.connect(self._on_root_flow_auto_toggled)
        root_flow_layout.addWidget(self.root_flow_spin)
        root_flow_layout.addWidget(self.root_flow_auto_cb)
        root_flow_unit = self._make_unit_label(self.tree_param_units.get('root_flow', ''))
        root_flow_layout.addWidget(root_flow_unit)
        root_flow_layout.addStretch()
        self.tree_param_unit_labels['root_flow'] = root_flow_unit
        tree_form.addRow(self._make_label_with_info("Root Flow:", _PARAM_DESCRIPTIONS['root_flow']), root_flow_layout)
        self.tree_param_widgets['root_flow'] = self.root_flow_spin

        self.max_nonconvex_spin = QSpinBox()
        self.max_nonconvex_spin.setRange(0, 1000000)
        self.max_nonconvex_spin.setSingleStep(10)
        self.max_nonconvex_spin.valueChanged.connect(lambda v: self._on_tree_param_changed('max_nonconvex_count', v))
        max_nonconvex_layout = QHBoxLayout()
        max_nonconvex_layout.setContentsMargins(0, 0, 0, 0)
        max_nonconvex_layout.setSpacing(6)
        max_nonconvex_layout.addWidget(self.max_nonconvex_spin)
        max_nonconvex_unit = self._make_unit_label(self.tree_param_units.get('max_nonconvex_count', ''))
        max_nonconvex_layout.addWidget(max_nonconvex_unit)
        max_nonconvex_layout.addStretch()
        self.tree_param_unit_labels['max_nonconvex_count'] = max_nonconvex_unit
        tree_form.addRow(self._make_label_with_info("Max Nonconvex Count:", _PARAM_DESCRIPTIONS['max_nonconvex_count']), max_nonconvex_layout)
        self.tree_param_widgets['max_nonconvex_count'] = self.max_nonconvex_spin

        scroll_layout.addWidget(tree_params_group)

        # Forest parameters (initially hidden)
        self.forest_params_group = QGroupBox("Forest Parameters")
        forest_form = QFormLayout()
        forest_form.setSpacing(10)
        forest_form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.forest_params_group.setLayout(forest_form)

        # Number of networks (Forest only)
        self.n_networks_spin = QSpinBox()
        self.n_networks_spin.setRange(1, 16)
        self.n_networks_spin.setValue(1)
        self.n_networks_spin.setToolTip("Number of independent vascular networks to generate in the forest")
        self.n_networks_spin.valueChanged.connect(self._on_forest_networks_changed)
        forest_form.addRow(self._make_label_with_info("Networks:", _PARAM_DESCRIPTIONS['n_networks']), self.n_networks_spin)

        # Trees per network container (Forest only) – one spinner per network
        self._n_trees_per_network = [2]  # default: 1 network with 2 trees
        self._trees_per_network_widget = QWidget()
        self._trees_per_network_layout = QVBoxLayout()
        self._trees_per_network_layout.setContentsMargins(0, 0, 0, 0)
        self._trees_per_network_layout.setSpacing(4)
        self._trees_per_network_widget.setLayout(self._trees_per_network_layout)
        self._trees_per_network_spins = []
        forest_form.addRow(self._make_label_with_info("Trees per Network:", _PARAM_DESCRIPTIONS['trees_per_network']), self._trees_per_network_widget)
        self._rebuild_trees_per_network_spins(1)

        # Vessels per network container (Forest only) – one spinner per network
        self._n_vessels_per_network = [self.n_vessels_spin.value()]
        self._vessels_per_network_widget = QWidget()
        self._vessels_per_network_layout = QVBoxLayout()
        self._vessels_per_network_layout.setContentsMargins(0, 0, 0, 0)
        self._vessels_per_network_layout.setSpacing(4)
        self._vessels_per_network_widget.setLayout(self._vessels_per_network_layout)
        self._vessels_per_network_spins = []
        forest_form.addRow(
            self._make_label_with_info("Vessels per Network:", _PARAM_DESCRIPTIONS['vessels_per_network']),
            self._vessels_per_network_widget
        )
        self._rebuild_vessels_per_network_spins(1)

        # Curve type used when connecting trees into networks
        self.curve_type_combo = QComboBox()
        # Display labels mapped to internal curve_type strings used by svv.forest.connect.curve.Curve
        self.curve_type_combo.addItem("Bezier", userData="Bezier")
        self.curve_type_combo.addItem("Catmull-Rom", userData="CatmullRom")
        self.curve_type_combo.addItem("NURBS", userData="NURBS")
        self.curve_type_combo.setCurrentIndex(0)
        self.curve_type_combo.setToolTip("Curve family used when connecting forest trees into networks")
        forest_form.addRow(self._make_label_with_info("Connection Curve:", _PARAM_DESCRIPTIONS['connection_curve']), self.curve_type_combo)

        self.compete_cb = QCheckBox("Enable Competition")
        self.compete_cb.setToolTip("Enable competition between trees for territory")
        forest_form.addRow(self._make_label_with_info("Competition:", _PARAM_DESCRIPTIONS['competition']), self.compete_cb)

        self.decay_probability_spin = QDoubleSpinBox()
        self.decay_probability_spin.setRange(0.0, 1.0)
        self.decay_probability_spin.setSingleStep(0.05)
        self.decay_probability_spin.setDecimals(2)
        self.decay_probability_spin.setValue(0.9)
        self.decay_probability_spin.setToolTip("Probability of decay for competitive growth (0-1)")
        forest_form.addRow(
            self._make_label_with_info("Decay Probability:", _PARAM_DESCRIPTIONS['decay_probability']),
            self.decay_probability_spin
        )

        scroll_layout.addWidget(self.forest_params_group)
        self.forest_params_group.hide()

        # Advanced parameters
        advanced_group = QGroupBox("Advanced Parameters")
        advanced_form = QFormLayout()
        advanced_form.setSpacing(10)
        advanced_form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        advanced_group.setLayout(advanced_form)

        self.random_seed_spin = QSpinBox()
        self.random_seed_spin.setRange(-1, 999999)
        self.random_seed_spin.setValue(-1)
        self.random_seed_spin.setSpecialValueText("Random")
        self.random_seed_spin.setToolTip("Set random seed for reproducible results (-1 = random)")
        advanced_form.addRow(
            self._make_label_with_info("Random Seed:", _PARAM_DESCRIPTIONS['random_seed']),
            self.random_seed_spin
        )

        # Preallocation size (rows) for tree data
        self.preallocation_spin = QSpinBox()
        self.preallocation_spin.setRange(1000, 20000000)
        self.preallocation_spin.setSingleStep(10000)
        self.preallocation_spin.setValue(int(1e5))
        self.preallocation_spin.setToolTip(
            "Maximum number of rows preallocated for each tree's data array.\n"
            "Lower values reduce memory usage but may trigger resize operations during growth."
        )
        advanced_form.addRow(
            self._make_label_with_info("Preallocation Size (rows):", _PARAM_DESCRIPTIONS['preallocation_size']),
            self.preallocation_spin
        )

        scroll_layout.addWidget(advanced_group)

        # Add stretch to push everything to top
        scroll_layout.addStretch()

        scroll.setWidget(scroll_content)
        layout.addWidget(scroll)

        # Initialize tree parameter widgets with defaults
        default_trees = 2
        if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
            try:
                default_trees = self.parent_gui.point_selector.tree_combo.count()
            except Exception:
                default_trees = 2
        self._sync_tree_override_options(1, [default_trees])
        self._apply_params_to_widgets(self.tree_base_params)
        self._refresh_unit_labels()

        # Ensure the initial mode (single tree vs forest) is reflected in the
        # point selector's Tree dropdown and parameter visibility.
        try:
            self._on_mode_changed(self.mode_combo.currentIndex())
        except Exception:
            pass

        # Action buttons
        button_layout = QVBoxLayout()
        button_layout.setSpacing(8)

        self.generate_btn = QPushButton("Generate Tree/Forest")
        self.generate_btn.clicked.connect(self._generate)
        self.generate_btn.setObjectName("primaryButton")
        self.generate_btn.setToolTip("Generate vascular tree or forest with current parameters")
        button_layout.addWidget(self.generate_btn)

        self.connect_now_btn = QPushButton("Connect Forest")
        self.connect_now_btn.clicked.connect(self._connect_existing_forest)
        self.connect_now_btn.setObjectName("primaryButton")
        self.connect_now_btn.setToolTip("Connect the currently generated forest into a network")
        self.connect_now_btn.hide()
        button_layout.addWidget(self.connect_now_btn)

        self.export_btn = QPushButton("Export Configuration")
        self.export_btn.clicked.connect(self._export_config)
        self.export_btn.setObjectName("secondaryButton")
        self.export_btn.setToolTip("Export current configuration to JSON file")
        button_layout.addWidget(self.export_btn)

        layout.addLayout(button_layout)

    # ---- Tree parameter helpers ----
    def _build_tree_param_units(self):
        """Return mapping of TreeParameters to their display units."""
        us = self.tree_unit_system
        return {
            'kinematic_viscosity': us.kinematic_viscosity.symbol,
            'fluid_density': us.density.symbol,
            'murray_exponent': 'unitless',
            'radius_exponent': 'unitless',
            'length_exponent': 'unitless',
            'terminal_pressure': us.pressure.symbol,
            'root_pressure': us.pressure.symbol,
            'terminal_flow': us.volumetric_flow.symbol,
            'root_flow': us.volumetric_flow.symbol,
            'max_nonconvex_count': 'count'
        }

    def _unit_system_summary(self):
        base = self.tree_unit_system.base
        return (
            f"Length: {base.length.symbol}, "
            f"Mass: {base.mass.symbol}, "
            f"Time: {base.time.symbol} "
            f"(Pressure: {self.tree_unit_system.pressure.symbol}, "
            f"Flow: {self.tree_unit_system.volumetric_flow.symbol})"
        )

    def _make_unit_label(self, text):
        label = QLabel(text or "")
        label.setStyleSheet("color: #6a6a6a;")
        return label

    def _make_info_button(self, description):
        """Create a small info-circle button that shows *description* on click."""
        btn = QToolButton()
        btn.setIcon(CADIcons.get_icon(
            "info",
            color=CADTheme.get_color('text', 'secondary'),
            size=14,
        ))
        btn.setIconSize(QSize(14, 14))
        btn.setFixedSize(18, 18)
        btn.setCursor(Qt.WhatsThisCursor)
        btn.setStyleSheet(
            "QToolButton { border: none; background: transparent; padding: 0px; }"
            "QToolButton:hover { background: rgba(255,255,255,0.08); border-radius: 9px; }"
        )
        btn.clicked.connect(
            lambda _=None, d=description, b=btn: QToolTip.showText(
                b.mapToGlobal(b.rect().bottomLeft()), d, b
            )
        )
        return btn

    def _make_label_with_info(self, text, description):
        """Return a QWidget containing a QLabel and a small info icon button."""
        widget = QWidget()
        lay = QHBoxLayout()
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(2)
        widget.setLayout(lay)
        lay.addWidget(QLabel(text))
        lay.addWidget(self._make_info_button(description))
        lay.addStretch()
        return widget

    def _refresh_unit_labels(self):
        self.tree_param_units = self._build_tree_param_units()
        for key, label in self.tree_param_unit_labels.items():
            if label is None:
                continue
            label.setText(self.tree_param_units.get(key, ''))

    def _default_tree_params(self, template=None):
        """Return default tree parameter values as a dict."""
        tp = template or TreeParameters(unit_system=self.tree_unit_system)
        return {
            'kinematic_viscosity': tp.kinematic_viscosity,
            'fluid_density': tp.fluid_density,
            'murray_exponent': tp.murray_exponent,
            'radius_exponent': tp.radius_exponent,
            'length_exponent': tp.length_exponent,
            'terminal_pressure': tp.terminal_pressure,
            'root_pressure': tp.root_pressure,
            'terminal_flow': tp.terminal_flow,
            'root_flow': tp.root_flow,
            'max_nonconvex_count': tp.max_nonconvex_count
        }

    def _on_unit_system_changed(self):
        """
        Handle changes to the base unit system from the dropdowns.

        This updates self.tree_unit_system and converts existing dimensional
        TreeParameters into the new units so values remain consistent.
        """
        length_symbol = self.length_unit_combo.currentText()
        mass_symbol = self.mass_unit_combo.currentText()
        time_symbol = self.time_unit_combo.currentText()

        # Build a new UnitSystem with selected base units
        new_unit_system = UnitSystem(length=length_symbol, mass=mass_symbol, time=time_symbol)

        # Update base params to use the new unit system, converting stored values
        tp = TreeParameters(unit_system=self.tree_unit_system)
        # Rehydrate tp from tree_base_params
        for key, value in self.tree_base_params.items():
            if value is None:
                continue
            tp.set(key, value)
        tp.set_unit_system(new_unit_system, convert_existing=True)

        # Update panel state
        self.tree_unit_system = new_unit_system
        self.tree_base_params = self._default_tree_params(tp)
        self._apply_params_to_widgets(self.tree_base_params)
        self._refresh_unit_labels()

        # Notify listeners that the length unit has changed
        self.length_unit_changed.emit(length_symbol)

    def _add_tree_param_spin(self, form, key, label, min_v, max_v, decimals, step, unit_text=None, info_text=None):
        """Create and register a QDoubleSpinBox for a tree parameter."""
        spin = QDoubleSpinBox()
        spin.setRange(min_v, max_v)
        spin.setDecimals(decimals)
        spin.setSingleStep(step)
        spin.valueChanged.connect(lambda v, k=key: self._on_tree_param_changed(k, v))
        widget = spin
        if unit_text is not None:
            row_widget = QWidget()
            row_layout = QHBoxLayout(row_widget)
            row_layout.setContentsMargins(0, 0, 0, 0)
            row_layout.setSpacing(6)
            row_layout.addWidget(spin)
            unit_label = self._make_unit_label(unit_text)
            row_layout.addWidget(unit_label)
            row_layout.addStretch()
            self.tree_param_unit_labels[key] = unit_label
            widget = row_widget
        row_label = (
            self._make_label_with_info(label + ":", info_text)
            if info_text else label + ":"
        )
        form.addRow(row_label, widget)
        self.tree_param_widgets[key] = spin

    def _apply_params_to_widgets(self, values):
        """Load parameter values into UI controls."""
        for key, widget in self.tree_param_widgets.items():
            if key not in values:
                continue
            with QSignalBlocker(widget):
                widget.setValue(values[key] if values[key] is not None else widget.minimum())
        if 'root_flow' in values:
            rf_val = values['root_flow']
            if rf_val is None:
                with QSignalBlocker(self.root_flow_auto_cb):
                    self.root_flow_auto_cb.setChecked(True)
                with QSignalBlocker(self.root_flow_spin):
                    self.root_flow_spin.setEnabled(False)
                    self.root_flow_spin.setValue(self.root_flow_spin.minimum())
            else:
                with QSignalBlocker(self.root_flow_auto_cb):
                    self.root_flow_auto_cb.setChecked(False)
                with QSignalBlocker(self.root_flow_spin):
                    self.root_flow_spin.setEnabled(True)
                    self.root_flow_spin.setValue(rf_val)
        if 'max_nonconvex_count' in values:
            with QSignalBlocker(self.max_nonconvex_spin):
                self.max_nonconvex_spin.setValue(values['max_nonconvex_count'])

    def _current_override_key(self):
        """Return (network, tree) when in forest mode, else None."""
        if self.mode_combo.currentIndex() != 1:  # forest index
            return None
        net = self.override_network_combo.currentIndex()
        tree = self.override_tree_combo.currentIndex()
        if net < 0 or tree < 0:
            return None
        return (net, tree)

    def _on_tree_override_changed(self, *_):
        """Reload widgets when a different tree override is selected."""
        # Ensure tree combo matches selected network's tree count if available
        if hasattr(self, "_trees_per_network"):
            net = self.override_network_combo.currentIndex()
            desired = 0
            if 0 <= net < len(self._trees_per_network):
                desired = self._trees_per_network[net]
            self._ensure_tree_combo_count(desired)
        self._load_tree_params_for_selected()

    def _load_tree_params_for_selected(self):
        """Apply the appropriate parameter set for the selected tree or base."""
        key = self._current_override_key()
        if key is None:
            self._apply_params_to_widgets(self.tree_base_params)
            return
        values = self.tree_param_overrides.get(key, self.tree_base_params)
        self._apply_params_to_widgets(values)

    def _on_root_flow_auto_toggled(self, state):
        """Handle toggling of root flow auto mode."""
        is_auto = state == Qt.Checked
        self.root_flow_spin.setEnabled(not is_auto)
        if is_auto:
            self._on_tree_param_changed('root_flow', None)
        else:
            self._on_tree_param_changed('root_flow', self.root_flow_spin.value())

    def _on_tree_param_changed(self, key, value):
        """Update stored parameters when a field changes."""
        if key == 'root_flow' and self.root_flow_auto_cb.isChecked():
            value = None
        if self.mode_combo.currentIndex() == 0:
            # Single tree / base params
            self.tree_base_params[key] = value
        else:
            tree_key = self._current_override_key()
            if tree_key is None:
                return
            if tree_key not in self.tree_param_overrides:
                # Start from base values
                self.tree_param_overrides[tree_key] = dict(self.tree_base_params)
            self.tree_param_overrides[tree_key][key] = value

    def _sync_tree_override_options(self, n_networks, n_trees_per_network):
        """Update the override dropdowns based on current forest layout."""
        self._trees_per_network = list(n_trees_per_network)
        with QSignalBlocker(self.override_network_combo):
            self.override_network_combo.clear()
            for i in range(n_networks):
                self.override_network_combo.addItem(f"Network {i}")
        with QSignalBlocker(self.override_tree_combo):
            self.override_tree_combo.clear()
            max_trees = max(n_trees_per_network) if n_trees_per_network else 0
            for j in range(max_trees):
                self.override_tree_combo.addItem(f"Tree {j}")
        if n_trees_per_network:
            self._ensure_tree_combo_count(n_trees_per_network[0])
        # Prune overrides that no longer fit
        valid_keys = set()
        for net_idx, n_t in enumerate(n_trees_per_network):
            for t_idx in range(n_t):
                valid_keys.add((net_idx, t_idx))
        self.tree_param_overrides = {
            k: v for k, v in self.tree_param_overrides.items() if k in valid_keys
        }
        self._load_tree_params_for_selected()

    def _ensure_tree_combo_count(self, count):
        """Ensure the tree combo has the desired number of entries."""
        with QSignalBlocker(self.override_tree_combo):
            current = self.override_tree_combo.count()
            if count < 0:
                count = 0
            if current != count:
                self.override_tree_combo.clear()
                for i in range(count):
                    self.override_tree_combo.addItem(f"Tree {i}")

    def _on_point_selector_network_changed(self, value):
        """Sync override combos when point selector network count changes."""
        # When the user changes the network count in the point selector, keep
        # override combos and per-network spinners in sync.
        if hasattr(self, "_trees_per_network_spins") and self.forest_params_group.isVisible():
            with QSignalBlocker(self.n_networks_spin):
                self.n_networks_spin.setValue(value)
            self._rebuild_trees_per_network_spins(value)
            self._rebuild_vessels_per_network_spins(value)
            trees_per_network = list(self._n_trees_per_network)
        else:
            try:
                n_trees = self.parent_gui.point_selector.tree_combo.count()
            except Exception:
                n_trees = 2
            trees_per_network = [n_trees] * value
        if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
            try:
                self.parent_gui.point_selector.set_trees_per_network(trees_per_network)
            except Exception:
                pass
        self._sync_tree_override_options(value, trees_per_network)

    def _build_tree_parameters(self, values):
        """Create a TreeParameters instance from a value dict."""
        tp = TreeParameters(unit_system=self.tree_unit_system)
        tp.kinematic_viscosity = values.get('kinematic_viscosity', tp.kinematic_viscosity)
        tp.fluid_density = values.get('fluid_density', tp.fluid_density)
        tp.murray_exponent = values.get('murray_exponent', tp.murray_exponent)
        tp.radius_exponent = values.get('radius_exponent', tp.radius_exponent)
        tp.length_exponent = values.get('length_exponent', tp.length_exponent)
        tp.terminal_pressure = values.get('terminal_pressure', tp.terminal_pressure)
        tp.root_pressure = values.get('root_pressure', tp.root_pressure)
        tp.terminal_flow = values.get('terminal_flow', tp.terminal_flow)
        tp.root_flow = values.get('root_flow', tp.root_flow)
        tp.max_nonconvex_count = values.get('max_nonconvex_count', tp.max_nonconvex_count)
        return tp

    # ---- Forest layout helpers ----
    def _on_forest_networks_changed(self, value):
        """
        Handle changes to the number of networks for forest generation.

        This rebuilds the per-network tree count spinners, updates the point
        selector's network dropdown, and syncs the per-tree override dropdowns.
        """
        if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
            ps = self.parent_gui.point_selector
            try:
                if hasattr(ps, "set_network_count"):
                    ps.set_network_count(value)
            except Exception:
                pass
        self._rebuild_trees_per_network_spins(value)
        self._rebuild_vessels_per_network_spins(value)
        if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
            try:
                self.parent_gui.point_selector.set_trees_per_network(list(self._n_trees_per_network))
            except Exception:
                pass
        self._sync_tree_override_options(value, list(self._n_trees_per_network))

    def _rebuild_trees_per_network_spins(self, n_networks):
        """Rebuild the per-network tree count spinners for the given number of networks."""
        old_values = list(self._n_trees_per_network)

        # Clear existing widgets
        while self._trees_per_network_layout.count():
            item = self._trees_per_network_layout.takeAt(0)
            w = item.widget()
            if w:
                w.deleteLater()
        self._trees_per_network_spins = []

        new_values = []
        for i in range(n_networks):
            default_val = old_values[i] if i < len(old_values) else 2
            new_values.append(default_val)

            row = QWidget()
            row_layout = QHBoxLayout()
            row_layout.setContentsMargins(0, 0, 0, 0)
            row_layout.setSpacing(4)
            row.setLayout(row_layout)

            if n_networks > 1:
                row_layout.addWidget(QLabel(f"Network {i}:"))

            spin = QSpinBox()
            spin.setRange(1, 16)
            spin.setValue(default_val)
            spin.setToolTip(f"Number of trees in network {i}")
            spin.valueChanged.connect(lambda v, idx=i: self._on_per_network_trees_changed(idx, v))
            row_layout.addWidget(spin)
            row_layout.addStretch()

            self._trees_per_network_layout.addWidget(row)
            self._trees_per_network_spins.append(spin)

        self._n_trees_per_network = new_values

    def _rebuild_vessels_per_network_spins(self, n_networks):
        """Rebuild the per-network vessel-count spinners for the given number of networks."""
        old_values = list(getattr(self, "_n_vessels_per_network", []))

        # Clear existing widgets
        while self._vessels_per_network_layout.count():
            item = self._vessels_per_network_layout.takeAt(0)
            w = item.widget()
            if w:
                w.deleteLater()
        self._vessels_per_network_spins = []

        default_base = int(self.n_vessels_spin.value()) if hasattr(self, "n_vessels_spin") else 100
        new_values = []
        for i in range(n_networks):
            default_val = int(old_values[i]) if i < len(old_values) else default_base
            default_val = max(1, default_val)
            new_values.append(default_val)

            row = QWidget()
            row_layout = QHBoxLayout()
            row_layout.setContentsMargins(0, 0, 0, 0)
            row_layout.setSpacing(4)
            row.setLayout(row_layout)

            if n_networks > 1:
                row_layout.addWidget(QLabel(f"Network {i}:"))

            spin = QSpinBox()
            spin.setRange(1, 100000)
            spin.setValue(default_val)
            spin.setToolTip(f"Number of vessel-growth steps for network {i}")
            spin.valueChanged.connect(lambda v, idx=i: self._on_per_network_vessels_changed(idx, v))
            row_layout.addWidget(spin)
            row_layout.addStretch()

            self._vessels_per_network_layout.addWidget(row)
            self._vessels_per_network_spins.append(spin)

        self._n_vessels_per_network = new_values

    def _on_per_network_vessels_changed(self, network_idx, value):
        """Handle a change to the vessel count for a specific network."""
        value = max(1, int(value))
        if network_idx < len(self._n_vessels_per_network):
            self._n_vessels_per_network[network_idx] = value

    def _on_per_network_trees_changed(self, network_idx, value):
        """Handle a change to the tree count for a specific network."""
        value = max(1, value)
        if network_idx < len(self._n_trees_per_network):
            self._n_trees_per_network[network_idx] = value

        # Update point selector
        if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
            try:
                self.parent_gui.point_selector.set_trees_per_network(list(self._n_trees_per_network))
            except Exception:
                pass

        # Sync override dropdowns
        n_networks = self.n_networks_spin.value() if hasattr(self, "n_networks_spin") else 1
        self._sync_tree_override_options(n_networks, list(self._n_trees_per_network))

    def _params_for_tree(self, net_idx, tree_idx):
        """Merge base params with per-tree override."""
        params = dict(self.tree_base_params)
        override = self.tree_param_overrides.get((net_idx, tree_idx))
        if override:
            params.update(override)
        return params

    def _on_mode_changed(self, index):
        """Handle mode selection change."""
        if index == 0:  # Single Tree
            self.forest_params_group.hide()
            self.tree_override_widget.hide()
            # When switching back to single tree, show base parameters
            self._apply_params_to_widgets(self.tree_base_params)
            # In single-tree mode, force a single network and a single tree
            if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
                ps = self.parent_gui.point_selector
                try:
                    if hasattr(ps, "set_network_count"):
                        ps.set_network_count(1)
                    ps.set_tree_count(1)
                except Exception:
                    pass
        else:  # Forest
            self.forest_params_group.show()
            self.tree_override_widget.show()
            # Sync override selectors with current point configuration
            if self.parent_gui and hasattr(self.parent_gui, "point_selector"):
                try:
                    n_networks = self.n_networks_spin.value()
                    trees_per_network = list(self._n_trees_per_network)
                    # Ensure list length matches current network count
                    while len(trees_per_network) < n_networks:
                        trees_per_network.append(2)
                    trees_per_network = trees_per_network[:n_networks]
                    # Update point selector's controls to reflect forest settings
                    ps = self.parent_gui.point_selector
                    if hasattr(ps, "set_network_count"):
                        ps.set_network_count(n_networks)
                    ps.set_trees_per_network(trees_per_network)
                    self._sync_tree_override_options(n_networks, trees_per_network)
                except Exception:
                    pass
            self._load_tree_params_for_selected()

        # Hide manual connect button until a forest is generated
        self._update_connect_button_visibility(force_hide=True)

    def _update_connect_button_visibility(self, force_hide=False):
        """
        Show or hide the manual forest connect button based on current state.

        Parameters
        ----------
        force_hide : bool
            If True, always hide the button regardless of current forest state.
        """
        if not hasattr(self, 'connect_now_btn'):
            return

        if force_hide:
            self.connect_now_btn.hide()
            self.connect_now_btn.setEnabled(True)
            return

        forest = self.parent_gui.forest if self.parent_gui else None
        should_show = forest is not None and getattr(forest, 'connections', None) is None
        self.connect_now_btn.setVisible(should_show)
        self.connect_now_btn.setEnabled(should_show)

    def _generate(self):
        """Generate Tree or Forest based on current configuration."""
        # Reset connect button until we know a fresh forest state
        self._update_connect_button_visibility(force_hide=True)

        if not self.parent_gui or not self.parent_gui.domain:
            QMessageBox.warning(
                self,
                "No Domain",
                "Please load a domain file before generating trees.\n\n"
                "Use File > Load Domain to get started."
            )
            self._record_telemetry(
                message="Generation requested without a loaded domain",
                level="warning",
                action="generate_without_domain",
            )
            return

        mode = self.mode_combo.currentIndex()
        n_vessels = self.n_vessels_spin.value()
        physical_clearance = self.physical_clearance_spin.value()
        convexity_tolerance = self.convexity_tolerance_spin.value()
        preallocation_step = self.preallocation_spin.value()
        random_seed = self.random_seed_spin.value() if self.random_seed_spin.value() >= 0 else None
        point_config = self.parent_gui.point_selector.get_configuration()

        if mode == 0:  # Single Tree
            if self.parent_gui:
                self.parent_gui.update_status(f"Generating tree with {n_vessels} vessels...")
            task_fn = lambda cancel, queue: self._generate_tree_task(
                n_vessels,
                physical_clearance,
                convexity_tolerance,
                preallocation_step,
                random_seed,
                point_config,
                cancel,
                queue
            )
            self._start_generation_task("Generating tree...", n_vessels, task_fn, self._on_tree_generated)
        else:  # Forest
            compete = self.compete_cb.isChecked()
            decay_prob = self.decay_probability_spin.value()
            n_networks = point_config.get('n_networks', 1)
            n_trees_per_network = point_config.get('n_trees_per_network', [2] * n_networks)
            self._sync_tree_override_options(n_networks, n_trees_per_network)
            per_network = list(getattr(self, "_n_vessels_per_network", []))
            if len(per_network) != n_networks:
                per_network = [int(n_vessels)] * int(n_networks)
            total_steps = max(per_network) if per_network else int(n_vessels)
            if self.parent_gui:
                if n_networks <= 4:
                    self.parent_gui.update_status(f"Generating forest (vessels/network: {per_network})...")
                else:
                    self.parent_gui.update_status(
                        f"Generating forest (vessels/network: min {min(per_network)} max {max(per_network)})..."
                    )
            task_fn = lambda cancel, queue: self._generate_forest_task(
                per_network,
                physical_clearance,
                convexity_tolerance,
                preallocation_step,
                random_seed,
                compete,
                decay_prob,
                point_config,
                cancel,
                queue
            )
            self._start_generation_task("Generating forest...", total_steps, task_fn, self._on_forest_generated)

    @staticmethod
    def _connect_forest_task(forest, curve_type="Bezier"):
        """
        Run forest.connect() in a worker context and return any exception.

        Returns
        -------
        Exception or None
            None on success, otherwise the raised exception.
        """
        try:
            forest.connect(curve_type=curve_type)
            return None
        except Exception as exc:  # Capture so GUI thread can handle it gracefully
            forest.connections = None
            # Report the exception to telemetry (if enabled) so background
            # connect failures are visible even when handled gracefully.
            try:
                try:
                    import sentry_sdk  # type: ignore[import]

                    with sentry_sdk.push_scope() as scope:
                        scope.set_tag("action", "forest_connect_worker")
                        sentry_sdk.capture_exception(exc)
                except Exception:
                    capture_exception(exc)
            except Exception:
                pass
            return exc

    def _connect_existing_forest(self):
        """Connect the currently generated forest if it is not already connected."""
        if not self.parent_gui or not getattr(self.parent_gui, 'forest', None):
            self._record_telemetry(
                message="Forest connect requested but no forest generated",
                level="warning",
                action="forest_connect_no_forest",
            )
            QMessageBox.information(
                self,
                "No Forest",
                "Generate a forest before attempting to connect it."
            )
            self._update_connect_button_visibility(force_hide=True)
            return

        forest = self.parent_gui.forest
        if getattr(forest, 'connections', None) is not None:
            self._record_telemetry(
                message="Forest connect requested but already connected",
                level="info",
                action="forest_connect_already_connected",
            )
            QMessageBox.information(
                self,
                "Already Connected",
                "The current forest is already connected."
            )
            self._update_connect_button_visibility()
            return

        # Run connection asynchronously to keep the UI responsive
        self._start_forest_connect(forest, show_success_dialog=True)

    def _start_forest_connect(self, forest, show_success_dialog=False):
        """
        Launch forest.connect() on a worker thread and show a non-blocking progress dialog.

        Parameters
        ----------
        forest : svv.forest.forest.Forest
            Forest object to connect
        show_success_dialog : bool
            If True, show an information dialog on success.
        """
        if self._connect_future is not None:
            return  # Connection already in progress

        # Record last action for crash diagnostics so that if the process is
        # killed during forest connection, the next GUI launch can report that
        # this was the last in-flight operation.
        try:
            settings = QSettings("svVascularize", "GUI")
            settings.setValue("session/last_action", "forest_connect")
        except Exception:
            pass

        if self.parent_gui and hasattr(self.parent_gui, "show_progress"):
            self.parent_gui.show_progress("Connecting forest...")
        if hasattr(self, 'connect_now_btn'):
            self.connect_now_btn.setEnabled(False)

        # Determine desired curve family from the GUI selector; default to Bezier
        curve_type = "Bezier"
        try:
            if hasattr(self, "curve_type_combo") and self.curve_type_combo is not None:
                data = self.curve_type_combo.currentData()
                curve_type = data if data is not None else self.curve_type_combo.currentText()
        except Exception:
            curve_type = "Bezier"

        future = self._executor.submit(self._connect_forest_task, forest, curve_type)
        self._connect_future = future
        self._poll_connect_future(future, forest, show_success_dialog)

    def _poll_connect_future(self, future, forest, show_success_dialog):
        """Poll a QtConcurrent future without blocking the UI."""
        if future.done():
            self._on_connect_finished(future, forest, show_success_dialog)
        else:
            QTimer.singleShot(50, lambda: self._poll_connect_future(future, forest, show_success_dialog))

    def _on_connect_finished(self, future, forest, show_success_dialog):
        """Handle completion of the background forest connection."""
        exc = None
        self._connect_future = None
        if self.parent_gui and hasattr(self.parent_gui, "hide_progress"):
            self.parent_gui.hide_progress()

        try:
            exc = future.result()
        except Exception as e:
            exc = e

        if exc:
            # Reset last_action now that connect has completed (even if it
            # failed gracefully inside Python). This ensures that only true
            # process-killing crashes during connect are reported on next run.
            try:
                settings = QSettings("svVascularize", "GUI")
                current = settings.value("session/last_action", "", type=str)
                if current == "forest_connect":
                    settings.setValue("session/last_action", "")
            except Exception:
                pass
            # Ensure connect failures are visible in telemetry even though the
            # GUI handles them with a warning dialog.
            tb = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
            self._record_telemetry(exc, action="forest_connect", traceback_str=tb)
            QMessageBox.warning(
                self,
                "Forest Connect Warning",
                f"Forest connection failed:\n\n{exc}"
            )
            if self.parent_gui:
                self.parent_gui.log_output(f"[Forest] connect failed: {exc}")
                self.parent_gui.update_status("Forest connection failed")
        else:
            if self.parent_gui:
                self.parent_gui.log_output("[Forest] connected")
                self.parent_gui.update_status("Forest connected")
                try:
                    self.parent_gui.vtk_widget.draw_forest(forest)
                except Exception:
                    # Keep UI responsive even if redraw fails
                    pass
            if show_success_dialog:
                QMessageBox.information(
                    self,
                    "Forest Connected",
                    "Forest connections created successfully."
                )

        if hasattr(self, 'connect_now_btn'):
            self.connect_now_btn.setEnabled(True)
        self._update_connect_button_visibility()

    # ---- Long-running task helpers ----
    def _start_generation_task(self, title, total_steps, task_fn, on_success):
        if self._generation_future is not None:
            self._record_telemetry(
                message="Generation requested while another generation is running",
                level="info",
                action="generation_already_running",
            )
            QMessageBox.information(self, "Task Running", "Please wait for the current operation to finish.")
            return
        if self._connect_future is not None:
            self._record_telemetry(
                message="Generation requested while forest connection is running",
                level="info",
                action="generation_connect_running",
            )
            QMessageBox.information(self, "Forest Connection Running", "Please wait for the forest connection to finish.")
            return

        self._generation_cancel_event = threading.Event()
        self._generation_progress_queue = Queue()
        self._generation_progress = QProgressDialog(title, "Cancel", 0, total_steps if total_steps is not None else 0, self)
        self._generation_progress.setWindowModality(Qt.WindowModal)
        self._generation_progress.setMinimumDuration(0)
        self._generation_progress.setAutoClose(True)
        if total_steps is None:
            self._generation_progress.setRange(0, 0)
        self._generation_progress.canceled.connect(self._generation_cancel_event.set)
        self._generation_progress.show()

        future = self._executor.submit(task_fn, self._generation_cancel_event, self._generation_progress_queue)
        self._generation_future = (future, on_success)
        QTimer.singleShot(50, self._poll_generation_future)

    def _poll_generation_future(self):
        if self._generation_future is None:
            return

        # Drain progress updates from the worker thread
        if self._generation_progress_queue is not None and self._generation_progress is not None:
            while not self._generation_progress_queue.empty():
                try:
                    value = self._generation_progress_queue.get_nowait()
                except Exception:
                    break
                try:
                    self._generation_progress.setValue(int(value))
                except Exception:
                    pass

        future, on_success = self._generation_future
        if future.done():
            self._finish_generation_future(on_success)
        else:
            QTimer.singleShot(100, self._poll_generation_future)

    def _finish_generation_future(self, on_success):
        future, _ = self._generation_future
        self._generation_future = None
        cancel_event = self._generation_cancel_event
        # Snapshot cancel state before closing the dialog. Some Qt styles can
        # emit `canceled` when the dialog is closed programmatically, which
        # would otherwise cause a successful run to be reported as canceled.
        was_canceled = bool(cancel_event and cancel_event.is_set())

        if self._generation_progress is not None:
            self._generation_progress.close()
            self._generation_progress = None
        self._generation_progress_queue = None
        self._generation_cancel_event = None

        try:
            result = future.result()
        except Exception as exc:
            # Log full traceback to the main window's output console so that
            # hard-to-reproduce worker errors (e.g., shape issues inside core
            # Tree/Forest routines) can be diagnosed with a complete stack trace.
            tb = traceback.format_exc()
            if self.parent_gui and hasattr(self.parent_gui, "log_output"):
                try:
                    self.parent_gui.log_output(tb)
                except Exception:
                    pass
            self._record_telemetry(exc, action="generation_worker", traceback_str=tb)
            QMessageBox.critical(self, "Generation Error", f"Failed to complete operation:\n\n{exc}")
            if self.parent_gui:
                self.parent_gui.update_status("Generation failed")
            return

        if result is None or was_canceled:
            if self.parent_gui:
                self.parent_gui.update_status("Generation canceled")
            return

        try:
            on_success(result)
        except Exception as exc:
            tb = traceback.format_exc()
            self._record_telemetry(exc, action="generation_finalize", traceback_str=tb)
            QMessageBox.critical(self, "Generation Error", f"Failed to finalize results:\n\n{exc}")
            if self.parent_gui:
                self.parent_gui.update_status("Generation failed")

    def _generate_tree_task(self, n_vessels, physical_clearance, convexity_tolerance,
                            preallocation_step, random_seed, config, cancel_event, progress_queue):
        """Generate a single tree in a worker thread."""
        from svv.tree.tree import Tree

        tree_params = self._build_tree_parameters(self.tree_base_params)
        tree = Tree(parameters=tree_params, preallocation_step=preallocation_step)
        tree.physical_clearance = physical_clearance
        if random_seed is not None:
            tree.random_seed = random_seed

        tree.set_domain(self.parent_gui.domain, convexity_tolerance)

        def _point_array(pt):
            if pt is None:
                return None
            arr = np.asarray(pt, dtype=float)
            d = self.parent_gui.domain.points.shape[1]
            # Accept either (d,) or (1,d) and normalize to a single
            # row of shape (1, d) so Domain.within/evaluate_fast always
            # see a 2D point array, avoiding scalar dists from cKDTree.
            if arr.ndim == 1 and arr.size == d:
                return arr.reshape(1, d)
            if arr.ndim == 2 and arr.shape[0] == 1 and arr.shape[1] == d:
                return arr
            raise ValueError(f"Invalid start point shape {arr.shape}; expected a single {d}-component coordinate.")

        def _vec_array(vec):
            if vec is None:
                return None
            arr = np.asarray(vec, dtype=float)
            d = self.parent_gui.domain.points.shape[1]
            if arr.ndim == 1 and arr.size == d:
                return arr
            if arr.ndim == 2 and arr.shape[0] == 1 and arr.shape[1] == d:
                return arr.reshape(-1)
            raise ValueError(f"Invalid direction shape {arr.shape}; expected {d}-component vector.")

        start_point = _point_array(config['start_points'][0][0])
        direction = _vec_array(config['directions'][0][0]) if config['directions'][0][0] is not None else None

        if start_point is not None:
            if direction is not None:
                tree.set_root(start_point, direction)
            else:
                tree.set_root(start_point)
        else:
            tree.set_root()

        update_every = max(1, n_vessels // 100)
        for i in range(n_vessels):
            if cancel_event.is_set():
                break
            tree.add()
            if (i + 1) % update_every == 0 and progress_queue is not None:
                progress_queue.put(i + 1)

        if progress_queue is not None:
            progress_queue.put(min(tree.n_terminals, n_vessels))

        return {
            'tree': tree,
            'physical_clearance': physical_clearance,
            'convexity_tolerance': convexity_tolerance,
            'canceled': cancel_event.is_set(),
        }

    def _generate_forest_task(self, n_vessels, physical_clearance, convexity_tolerance,
                        preallocation_step, random_seed, compete, decay_prob, config, cancel_event, progress_queue):
        """Generate a forest in a worker thread."""
        from svv.forest.forest import Forest

        n_networks = config['n_networks']
        n_trees_per_network = config['n_trees_per_network']
        start_points = config['start_points']
        directions = config['directions']

        def _point_array(pt):
            if pt is None:
                return None
            arr = np.asarray(pt, dtype=float).flatten()
            d = self.parent_gui.domain.points.shape[1]
            # Normalize forest start points to a single (1, d) row so
            # downstream Domain.within/evaluate_fast calls always see a
            # 2D array of points and never a 1D vector.
            # Handle case where we might have gotten multiple points - take first d values
            if arr.size >= d:
                arr = arr[:d]
            if arr.size == d:
                return arr.reshape(1, d)
            raise ValueError(f"Invalid start point: got {arr.size} values, expected {d}-component coordinate.")

        def _vec_array(vec):
            if vec is None:
                return None
            arr = np.asarray(vec, dtype=float).flatten()
            d = self.parent_gui.domain.points.shape[1]
            # Handle case where we might have gotten multiple directions - take first d values
            if arr.size >= d:
                arr = arr[:d]
            if arr.size == d:
                return arr
            raise ValueError(f"Invalid direction: got {arr.size} values, expected {d}-component vector.")

        norm_points = []
        norm_dirs = []
        for nets, dirs in zip(start_points, directions):
            p_net = []
            d_net = []
            for p, d in zip(nets, dirs):
                p_net.append(_point_array(p))
                d_net.append(_vec_array(d))
            norm_points.append(p_net)
            norm_dirs.append(d_net)
        start_points = norm_points
        directions = norm_dirs

        forest = Forest(
            domain=self.parent_gui.domain,
            n_networks=n_networks,
            n_trees_per_network=n_trees_per_network,
            start_points=start_points,
            directions=directions,
            physical_clearance=physical_clearance,
            compete=compete,
            preallocation_step=preallocation_step,
        )

        for net_idx, network in enumerate(forest.networks):
            for tree_idx, tree in enumerate(network):
                params = self._build_tree_parameters(self._params_for_tree(net_idx, tree_idx))
                tree.parameters = params
                if random_seed is not None:
                    tree.random_seed = random_seed

        forest.set_domain(self.parent_gui.domain, convexity_tolerance)
        forest.set_roots(start_points, directions)

        if isinstance(n_vessels, (list, tuple, np.ndarray)):
            per_network = [int(v) for v in np.asarray(n_vessels, dtype=int).reshape(-1).tolist()]
        else:
            per_network = [int(n_vessels)] * int(n_networks)
        if len(per_network) < int(n_networks):
            per_network.extend([per_network[-1] if per_network else 1] * (int(n_networks) - len(per_network)))
        per_network = per_network[:int(n_networks)]
        per_network = [max(1, int(v)) for v in per_network]

        total_steps = max(per_network) if per_network else 0
        update_every = max(1, total_steps // 100) if total_steps > 0 else 1
        for i in range(total_steps):
            if cancel_event.is_set():
                break
            active_networks = [idx for idx in range(int(n_networks)) if i < per_network[idx]]
            if not active_networks:
                break
            forest.add(1, network_id=active_networks, decay_probability=decay_prob)
            if (i + 1) % update_every == 0 and progress_queue is not None:
                progress_queue.put(i + 1)

        total_vessels = sum(tree.n_terminals for network in forest.networks for tree in network)
        if progress_queue is not None:
            progress_queue.put(total_steps)

        return {
            'forest': forest,
            'n_networks': n_networks,
            'n_vessels_per_network': per_network,
            'physical_clearance': physical_clearance,
            'convexity_tolerance': convexity_tolerance,
            'compete': compete,
            'decay_prob': decay_prob,
            'total_vessels': total_vessels,
            'canceled': cancel_event.is_set(),
        }

    def _on_tree_generated(self, result):
        tree = result.get('tree')
        if tree is None:
            if self.parent_gui:
                self.parent_gui.update_status("Tree generation canceled")
            return

        # Visualize and store
        self.parent_gui.vtk_widget.clear_trees()
        self.parent_gui.vtk_widget.clear_connections()
        # Group ID ('single', 0) used for visibility toggles
        self.parent_gui.vtk_widget.add_tree(tree, label="tree_single", group_id=("single", 0))
        self.parent_gui.trees = [tree]
        self.parent_gui.forest = None

        # Update Model Tree view
        if self.parent_gui and hasattr(self.parent_gui, "object_browser"):
            try:
                self.parent_gui.object_browser.set_single_tree(tree)
            except Exception:
                pass

        if self.parent_gui:
            self.parent_gui.update_status(f"Tree generated successfully with {tree.n_terminals} vessels")
            self.parent_gui.log_output(f"[Tree] done: {tree.n_terminals} vessels")

        QMessageBox.information(
            self,
            "Success",
            f"Tree generated successfully!\n\n"
            f"Total vessels: {tree.n_terminals}\n"
            f"Physical clearance: {result.get('physical_clearance')}\n"
            f"Convexity tolerance: {result.get('convexity_tolerance')}"
        )

    def _on_forest_generated(self, result):
        forest = result.get('forest')
        if forest is None:
            if self.parent_gui:
                self.parent_gui.update_status("Forest generation canceled")
            return

        self.parent_gui.vtk_widget.draw_forest(forest)
        self.parent_gui.forest = forest
        self.parent_gui.trees = []
        self._update_connect_button_visibility()

        # Update Model Tree with networks/trees layout
        if self.parent_gui and hasattr(self.parent_gui, "object_browser"):
            try:
                self.parent_gui.object_browser.set_forest(forest)
            except Exception:
                pass

        total_vessels = result.get('total_vessels', 0)
        n_networks = result.get('n_networks', 0)
        compete = result.get('compete', False)
        physical_clearance = result.get('physical_clearance')

        if self.parent_gui:
            self.parent_gui.update_status(
                f"Forest generated successfully with {total_vessels} vessels across {n_networks} networks"
            )
            self.parent_gui.log_output(f"[Forest] done: {total_vessels} vessels across {n_networks} networks")
            if forest.connections is None:
                self.parent_gui.log_output("[Forest] connect pending - click Connect Forest to link trees")

        QMessageBox.information(
            self,
            "Success",
            f"Forest generated successfully!\n\n"
            f"Total vessels: {total_vessels}\n"
            f"Networks: {n_networks}\n"
            f"Competition: {'Enabled' if compete else 'Disabled'}\n"
            f"Physical clearance: {physical_clearance}"
        )

    def _export_config(self):
        """Export the current configuration."""
        if self.parent_gui:
            self.parent_gui.save_configuration()

    def get_parameters(self):
        """
        Get the current parameter configuration.

        Returns
        -------
        dict
            Dictionary of parameters
        """
        params = {
            'mode': 'tree' if self.mode_combo.currentIndex() == 0 else 'forest',
            'n_vessels': self.n_vessels_spin.value(),
            'physical_clearance': self.physical_clearance_spin.value(),
            'convexity_tolerance': self.convexity_tolerance_spin.value(),
            'preallocation_step': self.preallocation_spin.value(),
            'random_seed': self.random_seed_spin.value() if self.random_seed_spin.value() >= 0 else None,
            'compete': self.compete_cb.isChecked(),
            'decay_probability': self.decay_probability_spin.value(),
            'tree_parameters': dict(self.tree_base_params),
            'tree_parameter_overrides': {str(k): v for k, v in self.tree_param_overrides.items()}
        }
        if self.mode_combo.currentIndex() != 0:
            params['n_vessels_per_network'] = list(getattr(self, "_n_vessels_per_network", []))
        return params



import numpy
import os
import tempfile
import multiprocessing as mp
import pyvista
import pymeshfix
from tqdm import trange, tqdm
from scipy.interpolate import splprep, splev
from scipy.spatial import cKDTree
from svv.utils.remeshing.remesh import remesh_surface, write_medit_sol
from svv.domain.routines.boolean import boolean


def find_unionable_pairs(lines, tubes, method='centerline', bbox_pad_factor=0.1, eps_factor=0.05):
    """
    Determine all tube index pairs (i, m) such that the initial point of the
    polyline `lines[i]` lies inside tube `tubes[m]`. These pairs are safe
    candidates for boolean unioning in a topology-aware manner (child into parent).

    Parameters
    ----------
    lines : list[pyvista.PolyData]
        Polylines with a single line cell each. Must have point array 'radius'.
    tubes : list[pyvista.PolyData]
        Triangulated surface meshes for each corresponding line (tube geometry).
    method : str, optional
        How to test containment. One of:
          - 'centerline' (default): project the line's initial point onto each
            candidate parent's polyline and compare distance to interpolated
            radius at the closest segment. Fast and robust to minor mesh issues.
          - 'surface': use VTK's select_enclosed_points against the tube surface
            to check if the point is inside. More expensive but purely geometric.
    bbox_pad_factor : float, optional
        Fraction of a tube's characteristic radius used to pad its axis-aligned
        bounding box for a quick inclusion prefilter. Default 0.1.
    eps_factor : float, optional
        Tolerance expressed as a fraction of the parent's median segment length
        for the centerline method. Default 0.05.

    Returns
    -------
    pairs : list[tuple[int, int]]
        List of (child_index, parent_index) pairs deemed unionable.

    Notes
    -----
    - This function assumes `len(lines) == len(tubes)` and that each line[i]
      corresponds to tube[i]. Pairs with identical indices (i == m) are skipped.
    - The 'centerline' method does not require watertight tubes and is typically
      much faster. The 'surface' method can be used when geometric certainty is
      preferred over speed.
    """
    import numpy as _np
    import pyvista as _pv

    assert len(lines) == len(tubes), "lines and tubes must have the same length"
    n = len(lines)

    # Precompute tube AABBs with padding and per-line helpers
    bounds = []
    pad = _np.zeros((n,), dtype=float)
    line_points = []
    line_radii = []
    line_seg_vecs = []
    line_seg_len2 = []
    line_seg_len = []
    line_med_seg = _np.zeros((n,), dtype=float)

    for i in range(n):
        l = lines[i]
        t = tubes[i]
        # Lines are expected to carry a 'radius' point array
        radii = _np.asarray(l['radius']).reshape(-1)
        pts = _np.asarray(l.points)
        line_points.append(pts)
        line_radii.append(radii)
        # Segment vectors and lengths for centerline projection
        if pts.shape[0] >= 2:
            vecs = pts[1:] - pts[:-1]
            seg_len2 = _np.einsum('ij,ij->i', vecs, vecs)
            seg_len = _np.sqrt(seg_len2)
            med = _np.median(seg_len) if seg_len.size > 0 else 0.0
        else:
            vecs = _np.zeros((0, 3))
            seg_len2 = _np.zeros((0,))
            seg_len = _np.zeros((0,))
            med = 0.0
        line_seg_vecs.append(vecs)
        line_seg_len2.append(seg_len2)
        line_seg_len.append(seg_len)
        line_med_seg[i] = med
        # Tube bounds and padding derived from characteristic radius
        b = t.bounds  # (xmin, xmax, ymin, ymax, zmin, zmax)
        bounds.append(b)
        r_char = float(_np.median(radii)) if radii.size else 0.0
        pad[i] = bbox_pad_factor * r_char

    def _in_padded_bounds(p, b, pad_val):
        return (b[0]-pad_val <= p[0] <= b[1]+pad_val and
                b[2]-pad_val <= p[1] <= b[3]+pad_val and
                b[4]-pad_val <= p[2] <= b[5]+pad_val)

    def _closest_distance_to_polyline(point, pts, vecs, seg_len2, radii):
        """Return (d_min, j_best, t_best, r_interp) for a point-to-polyline query."""
        if pts.shape[0] == 0:
            return _np.inf, -1, 0.0, 0.0
        if pts.shape[0] == 1:
            d = _np.linalg.norm(point - pts[0])
            return d, 0, 0.0, float(radii[0]) if radii.size else 0.0
        d_min = _np.inf
        j_best = -1
        t_best = 0.0
        r_best = 0.0
        p = point
        p1 = pts[:-1]
        v = vecs
        w = p - p1
        # Parametric projection along each segment
        with _np.errstate(invalid='ignore', divide='ignore'):
            t = _np.einsum('ij,ij->i', w, v) / seg_len2
        t = _np.clip(t, 0.0, 1.0, out=t)
        proj = p1 + (t[:, None] * v)
        diff = p - proj
        dists = _np.linalg.norm(diff, axis=1)
        if dists.size > 0:
            j = int(_np.argmin(dists))
            d_min = float(dists[j])
            j_best = j
            t_best = float(t[j])
            # Linear interpolate radius on segment
            if radii.size >= 2:
                r1 = float(radii[j])
                r2 = float(radii[j+1])
                r_best = (1.0 - t_best) * r1 + t_best * r2
            else:
                r_best = float(radii[0]) if radii.size else 0.0
        return d_min, j_best, t_best, r_best

    pairs = []
    # Build a single-point PolyData once for surface checks (reused per candidate)
    for i in range(n):
        # Initial point of child line i
        p0 = _np.asarray(lines[i].points[0])
        # Evaluate against all other tubes with quick AABB prefilter
        for m in range(n):
            if m == i:
                continue
            if not _in_padded_bounds(p0, bounds[m], pad[m]):
                continue
            if method == 'surface':
                # Use VTK select_enclosed_points
                pt_poly = _pv.PolyData(p0.reshape(1, 3))
                classified = tubes[m].select_enclosed_points(pt_poly, tolerance=0.0)
                # PyVista may name the array 'SelectedPoints' or 'Selected'
                if 'SelectedPoints' in classified.point_data:
                    inside = bool(int(classified.point_data['SelectedPoints'][0]))
                elif 'Selected' in classified.point_data:
                    inside = bool(int(classified.point_data['Selected'][0]))
                else:
                    inside = False
                if inside:
                    pairs.append((i, m))
            else:
                # Centerline projection test
                d_min, j_best, t_best, r_interp = _closest_distance_to_polyline(
                    p0, line_points[m], line_seg_vecs[m], line_seg_len2[m], line_radii[m]
                )
                # Tolerance scaled by parent's median segment length
                eps = eps_factor * (line_med_seg[m] if line_med_seg[m] > 0 else 1.0)
                if d_min <= (r_interp + eps):
                    pairs.append((i, m))

    return pairs

def get_longest_path(data, seed_edge):
    dig = True
    temp_edges = [seed_edge]
    while dig:
        keep_digging = []
        for edge in temp_edges:
            if not numpy.isnan(data[edge, 15]):
                temp_edges.extend([int(data[edge, 15])])
                if not numpy.isnan(data[edge, 16]):
                    temp_edges.extend([int(data[edge, 16])])
                temp_edges.remove(edge)
                keep_digging.append(True)
            else:
                keep_digging.append(False)
        dig = any(keep_digging)
    if len(temp_edges) == 1:
        return temp_edges
    edge_depths = []
    for edge in temp_edges:
        edge_depths.append(data[edge, 26])
    max_depth = max(edge_depths)
    max_edge_depths = [i for i, j in enumerate(edge_depths) if j == max_depth]
    paths = [[temp_edges[i]] for i in max_edge_depths]
    path_lengths = [data[i[0], 20] for i in paths]
    retrace = [True]*len(paths)
    while any(retrace):
        for i, path in enumerate(paths):
            if retrace[i]:
                path.insert(0, int(data[path[0], 17]))
                path_lengths[i] += data[path[0], 20]
                if paths[i][0] == seed_edge:
                    retrace[i] = False
    return paths[path_lengths.index(max(path_lengths))]


def get_alternate_path(data, seed_edge, reference=None):
    if reference is None:
        reference = get_longest_path(data, seed_edge)
    else:
        pass
    seed_edge_idx = reference.index(seed_edge)
    if seed_edge_idx == len(reference)-1:
        print('Seed edge is terminal and no alternative is possible. \n Computation finished.')
        return None
    else:
        children = [int(data[seed_edge, 15]), int(data[seed_edge, 16])]
        if children[0] not in reference:
            alternate_path = get_longest_path(data, children[0])
        else:
            alternate_path = get_longest_path(data, children[1])
    alternate_path.insert(0, seed_edge)
    return alternate_path


def get_branches(data):
    branches = []
    seed_edge = 0
    path = get_longest_path(data, seed_edge)
    branches.append(path)
    upcoming_evaluations = []
    upcoming_evaluations.extend(path[:-1])
    counter = [len(path[:-1])]
    idx = 0
    while len(upcoming_evaluations) > 0:
        if not numpy.isnan(data[upcoming_evaluations[-1], 15]) and not numpy.isnan(data[upcoming_evaluations[-1], 16]):
            pass
        else:
            upcoming_evaluations.pop(-1)
            counter[idx] -= 1
            if counter[idx] == 0:
                counter[idx] = None
                for i in reversed(range(len(counter))):
                    if counter[i] is not None:
                        idx = i
                        break
            continue
        path = get_alternate_path(data, upcoming_evaluations.pop(-1), reference=branches[idx])
        counter[idx] -= 1
        if counter[idx] == 0:
            counter[idx] = None
            for i in reversed(range(len(counter))):
                if counter[i] is not None:
                    idx = i
                    break
        branches.append(path)
        if len(path) > 2:
            upcoming_evaluations.extend(path[1:-1])
            counter.append(len(path[1:-1]))
            idx = len(counter) - 1
        else:
            counter.append(None)
    return branches


def get_points(data, branches):
    path_points = []
    primed = False
    for path in branches:
        branch_points = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_points.insert(0, data[edge, 3:6].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                branch_points.insert(0, data[edge, 3:6].tolist())
                three_fourths = data[edge, 0:3]*(1/4) + data[edge, 3:6]*(3/4)
                branch_points.insert(0, three_fourths.tolist())
                mid_point = (data[edge, 0:3] + data[edge, 3:6])/2
                branch_points.insert(0, mid_point.tolist())
                one_fourth = data[edge, 0:3]*(3/4) + data[edge, 3:6]*(1/4)
                branch_points.insert(0, one_fourth.tolist())
                branch_points.insert(0, data[edge, 0:3].tolist())
                primed = True
            elif len(branch_points) == 0:
                branch_points.insert(0, data[edge, 3:6].tolist())
                three_fourths = data[edge, 0:3]*(1/4) + data[edge, 3:6]*(3/4)
                branch_points.insert(0, three_fourths.tolist())
                mid_point = (data[edge, 0:3] + data[edge, 3:6])/2
                branch_points.insert(0, mid_point.tolist())
                one_fourth = data[edge, 0:3]*(3/4) + data[edge, 3:6]*(1/4)
                branch_points.insert(0, one_fourth.tolist())
            else:
                branch_points.insert(0, data[edge, 3:6].tolist())
                three_fourths = data[edge, 0:3]*(1/4) + data[edge, 3:6]*(3/4)
                branch_points.insert(0, three_fourths.tolist())
                mid_point = (data[edge, 0:3] + data[edge, 3:6])/2
                branch_points.insert(0, mid_point.tolist())
                one_fourth = data[edge, 0:3]*(3/4) + data[edge, 3:6]*(1/4)
                branch_points.insert(0, one_fourth.tolist())
        path_points.append(branch_points)
    return path_points


def get_radii(data, branches):
    path_radii = []
    primed = False
    for path in branches:
        branch_radii = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_radii.insert(0, data[path[1], 21])
            elif edge == path[0] and edge == 0 and not primed:
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                primed = True
            elif len(branch_radii) == 0:
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
            else:
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
                branch_radii.insert(0, data[edge, 21])
        path_radii.append(branch_radii)
    return path_radii


def get_normals(data, branches):
    path_normals = []
    primed = False
    for path in branches:
        branch_normals = []
        for edge in reversed(path):
            if edge == path[0] and primed:
                branch_normals.insert(0, data[path[1], 12:15].tolist())
            elif edge == path[0] and edge == 0 and not primed:
                vector_1 = data[edge, 12:15]
                mid_vector = (vector_1 + vector_2)/2
                branch_normals.insert(0, mid_vector.tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                primed = True
            elif len(branch_normals) == 0:
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                vector_2 = data[edge, 12:15]
            else:
                vector_1 = data[edge, 12:15]
                mid_vector = (vector_1+vector_2)/2
                branch_normals.insert(0, mid_vector.tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                branch_normals.insert(0, data[edge, 12:15].tolist())
                vector_2 = data[edge, 12:15]
        path_normals.append(branch_normals)
    return path_normals


def get_interpolated_sv_data(data):
    branches = get_branches(data)
    points = get_points(data, branches)
    radii = get_radii(data, branches)
    normals = get_normals(data, branches)
    path_frames = []
    for idx in range(len(branches)):
        frames = []
        for jdx in range(len(points[idx])):
            frame = []
            frame.extend(points[idx][jdx])
            frame.append(radii[idx][jdx])
            frame.extend(normals[idx][jdx])
            frames.append(frame)
        path_frames.append(frames)
    interp_xyz = []
    interp_r = []
    interp_n = []
    interp_xyzr = []
    for idx in range(len(branches)):
        p = numpy.array(points[idx]).T
        r = numpy.array(radii[idx]).T
        n = numpy.array(normals[idx]).T
        if len(points[idx]) == 2:
            interp_xyz.append(splprep(p, k=1, s=0))
            rr = numpy.vstack((interp_xyz[-1][1], r))
            interp_r.append(splprep(rr, k=1, s=0))
            xyzr = numpy.vstack((p, r))
            interp_xyzr.append(splprep(xyzr, k=1, s=0))
            #interp_n.append(splprep(n, k=1, s=0))
        elif len(points[idx]) == 3:
            interp_xyz.append(splprep(p, k=2, s=0))
            rr = numpy.vstack((interp_xyz[-1][1], r))
            interp_r.append(splprep(rr, k=1, s=0))
            xyzr = numpy.vstack((p, r))
            interp_xyzr.append(splprep(xyzr, k=2, s=0))
            #interp_n.append(splprep(n, k=2, s=0))
        else:
            interp_xyz.append(splprep(p, s=0))
            rr = numpy.vstack((interp_xyz[-1][1], r))
            interp_r.append(splprep(rr, k=1, s=0))
            xyzr = numpy.vstack((p, r))
            interp_xyzr.append(splprep(xyzr, s=0))
            #interp_n.append(splprep(n, s=0))
    return interp_xyz, interp_r, interp_n, path_frames, branches, interp_xyzr


def polyline_from_points(pts, r):
    poly = pyvista.PolyData()
    poly.points = pts
    cell = numpy.arange(0, len(pts), dtype=numpy.int_)
    cell = numpy.insert(cell, 0, len(pts))
    poly.lines = cell
    poly['radius'] = r
    return poly


def generate_polylines(xyz, r, num=1000):
    """
    This function generates Pyvista polydata polyline objects from a given set of spline interpolated
    data.

    Parameters
    ----------
    xyz : list
        A list of tuples containing the vector of knots, B-spline coefficients, and the degree of the N-degree
        curve spline
    r : list
        A list of tuples containing the vector of knots, B-spline coefficients, and the degree of the 1-D radial
        spline
    num : int
        The number of points to sample from the spline.

    Returns
    -------
    polylines : list
        A list of Pyvista polydata polyline objects.
    """
    polylines = []
    t = numpy.linspace(0, 1, num)
    for xyz_i, r_i in zip(xyz, r):
        x, y, z = splev(t, xyz_i[0])
        _, radius = splev(t, r_i[0])
        points = numpy.zeros((num, 3))
        points[:, 0] = x
        points[:, 1] = y
        points[:, 2] = z
        polylines.append(polyline_from_points(points, radius))
    return polylines


def generate_tube(polyline, hsize=None, radius_based=True):
    """
    This function generates a tube from a given polyline representing a single vessel of a
    vascular object (tree or forest).

    Parameters
    ----------
    polyline : pyvista.PolyData
        A polyline object representing the centerline of a vessel.
    hsize : float
        The mesh element size for the surface mesh of the vessel. When
        radius_based is True, this acts as the target size at the minimum
        centerline radius and scales proportionally elsewhere.
    radius_based : bool
        If True, writes a per-vertex sizing function (in.sol) proportional to
        the local centerline radius so MMG adapts edge sizes accordingly.

    Returns
    -------
    tube : pyvista.PolyData
        A triangulated tube surface mesh object representing the vessel.
    """
    tube = polyline.tube(radius=min(polyline['radius']), scalars='radius',
                         radius_factor=max(polyline['radius'])/min(polyline['radius']), capping=True)
    tube = tube.triangulate()
    tube = tube.compute_normals(auto_orient_normals=True)
    fix = pymeshfix.MeshFix(tube)
    fix.repair()
    tube = fix.mesh
    tube = tube.compute_normals(auto_orient_normals=True)
    if isinstance(hsize, type(None)):
        hsize = (min(polyline['radius'])*2*numpy.pi)/25
    if radius_based:
        # Per-vertex sizing based on local centerline radius using k-NN from KDTree.
        # Scale so hsize matches the size at the minimum radius; elsewhere scales proportionally.
        poly_pts = numpy.asarray(polyline.points, dtype=float)
        poly_rad = numpy.asarray(polyline['radius'], dtype=float).reshape(-1)
        rmin = float(poly_rad.min()) if poly_rad.size else 1.0
        scale = float(hsize) / rmin if rmin > 0 else float(hsize)

        surf_pts = numpy.asarray(tube.points, dtype=float)
        n_poly = poly_pts.shape[0]
        k_nn = min(4, n_poly) if n_poly > 0 else 1
        if n_poly == 0:
            sizes = numpy.full(surf_pts.shape[0], float(hsize), dtype=float)
        else:
            tree = cKDTree(poly_pts)
            d, idx = tree.query(surf_pts, k=k_nn)
            # Handle k=1 return shape
            if k_nn == 1:
                r_local = poly_rad[numpy.asarray(idx).reshape(-1)].astype(float)
            else:
                d = numpy.asarray(d, dtype=float)
                idx = numpy.asarray(idx, dtype=int)
                # Inverse-distance weights; protect against zero distances
                w = 1.0 / (d + 1e-12)
                w /= w.sum(axis=1, keepdims=True)
                r_neighbors = poly_rad[idx]
                r_local = (w * r_neighbors).sum(axis=1)
            sizes = scale * r_local
        # Ensure strictly positive sizes for MMG
        sizes = numpy.maximum(sizes, numpy.finfo(float).eps)
        tube.point_data['MeshSizingFunction'] = sizes
        # Write MMG sizing file the remesher will pick up in this temp directory
        write_medit_sol(tube, 'in.sol', array_name='MeshSizingFunction', scale=1, default_size=hsize)
    # If using a sizing function, let MMG drive sizes solely from in.sol (omit -hsiz)
    tube = remesh_surface(tube, hsiz=(None if radius_based else hsize), verbosity=0)
    tube = tube.compute_normals(auto_orient_normals=True)
    fix = pymeshfix.MeshFix(tube)
    fix.repair()
    tube = fix.mesh
    tube = tube.compute_normals(auto_orient_normals=True)
    return tube


def generate_tubes(polylines, hsize=None):
    """
    This function generates a list of tube surface meshes from a list of polyline objects representing
    the vessels of a vascular object (tree or forest).

    Parameters
    ----------
    polylines : list
        A list of Pyvista polyline objects representing the centerlines of the vessels.
    hsize : float
        The mesh element size for the surface meshes of the vessels.

    Returns
    -------
    tubes : list
        A list of Pyvista surface mesh objects representing the vessels.
    """
    tubes = []
    for polyline in polylines:
        tube = generate_tube(polyline, hsize=hsize)
        tubes.append(tube)
    return tubes


def _tube_worker(args):
    """Worker that builds a single tube in an isolated temp directory.

    Args
    ----
    args : tuple
        (idx, points, radius, hsize, radius_based)

    Returns
    -------
    tuple
        (idx, points, faces) for reconstructed PolyData.
    """
    idx, pts, rad, hsize, radius_based = args
    old_cwd = os.getcwd()
    # Isolate MMG temp files to avoid collisions between processes
    # As in the tetrahedralizer, prefer TEMP/TMP on Windows to avoid
    # issues when TMPDIR is set to an invalid POSIX-style path.
    tmp_root = None
    if os.name == "nt":
        for env_var in ("TEMP", "TMP"):
            candidate = os.environ.get(env_var)
            if candidate and os.path.isdir(candidate):
                tmp_root = candidate
                break

    with tempfile.TemporaryDirectory(prefix="svv_remesh_", dir=tmp_root) as tmpdir:
        try:
            os.chdir(tmpdir)
            poly = polyline_from_points(numpy.asarray(pts), numpy.asarray(rad))
            tube = generate_tube(poly, hsize=hsize, radius_based=radius_based)
            # Return minimal geometry to avoid pickling VTK objects
            return idx, numpy.asarray(tube.points), numpy.asarray(tube.faces)
        finally:
            os.chdir(old_cwd)


def generate_tubes_parallel(polylines, hsize=None, processes=None, chunksize=1, start_method=None, show_progress=True, radius_based=False):
    """
    Parallel tube generation using multiprocessing. Each tube is built in a
    separate process and in a per-process temporary directory to avoid MMG
    temp-file collisions.

    Parameters
    ----------
    polylines : list[pyvista.PolyData]
        Centerline polylines with point-data array 'radius'.
    hsize : float, optional
        Target surface edge size for remeshing (forwarded to generate_tube).
    processes : int, optional
        Number of worker processes. Defaults to `os.cpu_count()`.
    chunksize : int, optional
        Chunk size for Pool.imap. Default 1.
    start_method : {"spawn","fork","forkserver"}, optional
        Multiprocessing start method. Defaults to 'spawn' when available.
    show_progress : bool, optional
        If True, display a tqdm progress bar.
    radius_based : bool, optional
        If True, build a per-vertex MMG sizing function based on the local
        centerline radius and pass it via in.sol, yielding radius-proportional
        edge sizes. Default False.

    Returns
    -------
    list[pyvista.PolyData]
        Reconstructed tube meshes in the same order as input polylines.
    """
    n = len(polylines)
    if n == 0:
        return []

    # Optional sequential fallback, useful on environments where
    # multiprocessing with the 'spawn' context is problematic
    # (e.g., inline scripts on Windows such as `python - <<` in CI).
    disable_env = os.environ.get("SVV_DISABLE_TUBE_PARALLEL", "")
    disable_parallel = disable_env.strip().lower() in {"1", "true", "yes", "on"}
    if disable_parallel:
        tubes = []
        iterable = polylines
        if show_progress:
            iterable = tqdm(iterable, total=n, desc='Generate tubes ', unit='tube', leave=False)
        for pl in iterable:
            tube = generate_tube(pl, hsize=hsize, radius_based=radius_based)
            tubes.append(tube)
        return tubes

    # Serialize inputs (avoid sending VTK objects across processes)
    tasks = []
    for i, pl in enumerate(polylines):
        pts = numpy.asarray(pl.points)
        if 'radius' not in pl.point_data:
            raise KeyError("Each polyline must have a 'radius' point-data array")
        rad = numpy.asarray(pl['radius']).reshape(-1)
        tasks.append((i, pts, rad, hsize, radius_based))

    # Choose a safe start method to avoid forking VTK state
    if start_method is None:
        try:
            ctx = mp.get_context('spawn')
        except ValueError:
            ctx = mp.get_context()
    else:
        ctx = mp.get_context(start_method)

    tubes_arrays = [None] * n
    with ctx.Pool(processes=processes) as pool:
        iterator = pool.imap(_tube_worker, tasks, chunksize)
        if show_progress:
            iterator = tqdm(iterator, total=n, desc='Generate tubes ', unit='tube', leave=False)
        for idx, pts, faces in iterator:
            tubes_arrays[idx] = (pts, faces)

    # Reconstruct PolyData objects in parent process
    tubes = []
    for i in range(n):
        pts, faces = tubes_arrays[i]
        tube = pyvista.PolyData(pts, faces)
        # Ensure normals are available
        tube = tube.compute_normals(auto_orient_normals=True)
        tubes.append(tube)
    return tubes

def union_tubes(tubes, lines, cap_resolution=40):
    """
    This function performs iterative boolean union operations on a list of Pyvista polydata surface
    meshes to build one single surface mesh representing the entire vascular object.

    Parameters
    ----------
    tubes : list
        A list of Pyvista polydata objects representing the tubes of the vessels in a vascular object.

    Returns
    -------
    union : pyvista.PolyData
        The result of the boolean union operation.
    """
    model = boolean(tubes[0], tubes[1], operation='union')
    model = model.compute_normals(auto_orient_normals=True)
    hsize = (min(min(lines[0]['radius']), min(lines[1]['radius']))*2*numpy.pi)/cap_resolution
    model = remesh_surface(model, hsiz=hsize)
    model = model.compute_normals(auto_orient_normals=True)
    if len(tubes) > 2:
        for i in range(2, len(tubes)):
            model = boolean(model, tubes[i], operation='union')
            model = model.compute_normals(auto_orient_normals=True)
            hsize = min(hsize, (min(lines[i]['radius'])*2*numpy.pi)/cap_resolution)
            model = remesh_surface(model, hsiz=hsize)
            model = model.compute_normals(auto_orient_normals=True)
    model.cell_data['hsize'] = 0
    model.cell_data['hsize'][0] = hsize
    return model


def build_watertight_solid(tree, cap_resolution=40):
    """
    This function builds a solid surface mesh from a given vascular tree object.
    This mesh should be guaranteed to be watertight and define a closed manifold.

    Parameters
    ----------
    tree : svtoolkit.tree.tree.Tree
        A vascular tree object.

    Returns
    -------
    model : pyvista.PolyData
        A solid surface mesh representing the entire vascular tree.
    """
    xyz, r, _, _, branches, _ = get_interpolated_sv_data(tree.data)
    lines = generate_polylines(xyz, r)
    tubes = generate_tubes(lines)
    model = union_tubes(tubes, lines, cap_resolution=cap_resolution)
    #tubes = generate_tubes_parallel(lines, radius_based=True)
    #model = union_tubes_balanced(tubes, lines, cap_resolution=cap_resolution)
    # Remove poor quality elements and repair the mesh.
    cell_quality = model.compute_cell_quality(quality_measure='scaled_jacobian')
    keep = cell_quality.cell_data["CellQuality"] > 0.1
    if not numpy.all(keep):
        print("Removing poor quality elements from the mesh.")
        #keep = numpy.argwhere(keep).flatten()
        #non_manifold_model = model.extract_cells(keep)
        #non_manifold_model = non_manifold_model.extract_surface()
        fix = pymeshfix.MeshFix(model) # non_manifold_model)
        fix.repair(verbose=True)
        hsize = model.cell_data["hsize"][0] #hsize
        model = fix.mesh.compute_normals(auto_orient_normals=True)
        #model.hsize = hsize
        model.cell_data['hsize'] = 0
    model.cell_data['hsize'][0] = hsize
    return model


def union_tubes_balanced(tubes, lines, cap_resolution=40, method='centerline', engine='manifold', fix_mesh=True):
    """
    Perform topology-aware, compute-balanced boolean unions over a set of vessel tubes.

    The function:
    - Detects unionable parent/child pairs using `find_unionable_pairs` (child's
      initial point inside parent's tube).
    - Schedules unions with a min-heap keyed by current mesh sizes so that the
      smallest unions occur first, keeping intermediates compact.
    - Updates the graph as components merge until each connected component is
      reduced to a single mesh.
    - Applies a final remesh and normal computation once on the result.

    Parameters
    ----------
    tubes : list[pyvista.PolyData]
        Per-vessel tube surface meshes (triangulated).
    lines : list[pyvista.PolyData]
        Per-vessel polylines carrying point-data array 'radius'.
    cap_resolution : int, optional
        Controls target edge size for final remeshing via: hsiz ≈ 2π·min_radius / cap_resolution.
        Default 40.
    method : {'centerline','surface'}, optional
        Containment test used by `find_unionable_pairs`. Default 'centerline' (fast).
    engine : str, optional
        Boolean engine forwarded to trimesh via `boolean`. Default 'manifold'.
    fix_mesh : bool, optional
        Whether to run mesh fixing during boolean operations. Forwarded to `boolean`.
        Default True.

    Returns
    -------
    model : pyvista.PolyData
        The watertight surface representing the union of all tubes, remeshed once
        at a global target size derived from cap_resolution.
    """
    import heapq
    import numpy as _np
    import pyvista as _pv

    n = len(tubes)
    if n == 0:
        return _pv.PolyData()
    if len(lines) != n:
        raise ValueError("lines and tubes must have the same length")

    # Determine all unionable pairs (child, parent)
    pairs = find_unionable_pairs(lines, tubes, method=method)
    if len(pairs) == 0:
        # Fallback: sequential union as last resort
        return union_tubes(tubes, lines, cap_resolution=cap_resolution)

    # Disjoint-set (union-find)
    parent = list(range(n))
    rank = [0] * n

    def find(x):
        while parent[x] != x:
            parent[x] = parent[parent[x]]
            x = parent[x]
        return x

    def union(a, b):
        ra, rb = find(a), find(b)
        if ra == rb:
            return ra
        if rank[ra] < rank[rb]:
            parent[ra] = rb
            return rb
        elif rank[ra] > rank[rb]:
            parent[rb] = ra
            return ra
        else:
            parent[rb] = ra
            rank[ra] += 1
            return ra

    # Component meshes and sizes
    comp_mesh = {i: tubes[i] for i in range(n)}
    comp_size = {i: int(tubes[i].n_cells) for i in range(n)}

    # Build neighbor sets from pairs (treat undirected for scheduling)
    neighbors = {i: set() for i in range(n)}
    for i, m in pairs:
        if i == m:
            continue
        neighbors[i].add(m)
        neighbors[m].add(i)

    # Min-heap of candidate unions keyed by sum of component sizes
    heap = []
    for i, m in pairs:
        ra, rb = find(i), find(m)
        if ra == rb:
            continue
        key = comp_size[ra] + comp_size[rb]
        heapq.heappush(heap, (key, ra, rb))

    # Compute total planned unions for progress: sum(|C|-1) over connected components
    # Build node set from pairs and traverse via neighbors
    nodes = set()
    for i, m in pairs:
        nodes.add(i); nodes.add(m)
    visited = set()
    total_merges = 0
    for v in nodes:
        if v in visited:
            continue
        # BFS/DFS over neighbor graph
        stack = [v]
        comp_count = 0
        while stack:
            u = stack.pop()
            if u in visited:
                continue
            visited.add(u)
            comp_count += 1
            for w in neighbors.get(u, ()):  # undirected traversal
                if w not in visited:
                    stack.append(w)
        if comp_count > 0:
            total_merges += max(comp_count - 1, 0)

    pbar = tqdm(total=total_merges, desc='Union tubes ', unit='union', leave=False)

    # Iteratively union smallest available pairs until components are reduced
    while heap:
        _, a0, b0 = heapq.heappop(heap)
        ra, rb = find(a0), find(b0)
        if ra == rb:
            continue  # stale
        # Perform boolean union between current component meshes
        try:
            merged = boolean(comp_mesh[ra], comp_mesh[rb], operation='union', fix_mesh=fix_mesh, engine=engine)
        except Exception as e:
            # If an error occurs, try with fix_mesh=True as a fallback
            if not fix_mesh:
                merged = boolean(comp_mesh[ra], comp_mesh[rb], operation='union', fix_mesh=True, engine=engine)
            else:
                raise e
        r = union(ra, rb)
        comp_mesh[r] = merged
        comp_size[r] = int(merged.n_cells)
        pbar.update(1)

        # Merge neighbor sets and push new edges against r
        na = neighbors.get(ra, set())
        nb = neighbors.get(rb, set())
        merged_neighbors = set()
        for c in (na | nb):
            rc = find(c)
            if rc != r:
                merged_neighbors.add(rc)
        neighbors[r] = merged_neighbors

        for rc in merged_neighbors:
            key = comp_size[r] + comp_size[find(rc)]
            heapq.heappush(heap, (key, r, rc))

    pbar.close()

    # Collect final components (roots)
    roots = {}
    for i in range(n):
        ri = find(i)
        roots[ri] = comp_mesh[ri]

    # Merge disjoint components without boolean (safe if disjoint)
    if len(roots) == 1:
        model = next(iter(roots.values()))
    else:
        model = _pv.merge(list(roots.values()))

    # Final remesh once using a global target size derived from min radius
    try:
        min_r = min(float(_np.min(lines[i]['radius'])) for i in range(len(lines)) if lines[i].n_points > 0)
        hsize = (2.0 * _np.pi * min_r) / float(cap_resolution)
    except Exception:
        # Fallback heuristic if radii missing
        hsize = None

    #if hsize is not None and hsize > 0:
    #    model = remesh_surface(model, hsiz=hsize, verbosity=0)
    model = model.compute_normals(auto_orient_normals=True)
    #fix = pymeshfix.MeshFix(model)
    #fix.repair()
    #model = fix.mesh
    if hsize is not None:
        model.cell_data['hsize'] = 0
        model.cell_data['hsize'][0] = hsize
    return model


def build_merged_solid(tree):
    """
    This function builds a solid surface mesh from a given vascular tree object.
    This mesh is not guaranteed to be watertight or define a closed manifold.
    Instead, this function may be useful in visualizing tree models and displaying
    reduced-order fluid flow simulations in an efficient manner.

    Parameters
    ----------
    tree : svtoolkit.tree.tree.Tree
        A vascular tree object.

    Returns
    -------
    merged_model : pyvista.PolyData
        A merged surface mesh representing the entire vascular tree.
    """
    vessels = []
    for i in trange(tree.data.shape[0], desc="Building Merged Solid"):
        center = (tree.data[i, 0:3] + tree.data[i, 3:6]) / 2
        direction = tree.data.get('w_basis', i)
        radius = tree.data.get('radius', i)
        length = tree.data.get('length', i)
        assert length > 0, "Length must be positive"
        assert radius > 0, "Radius must be positive"
        vessels.append(pyvista.Cylinder(center=center, direction=direction, radius=radius, height=length))
    merged_model = pyvista.merge(vessels)
    return merged_model



from types import SimpleNamespace

import numpy as np
import pytest

from svv.forest.export.export_spline import write_splines as write_forest_splines
from svv.visualize.spline_export import export_spline_files


def _make_tree(x_offset: float, y_offset: float = 0.0, radius: float = 0.1):
    data = np.full((2, 31), np.nan)

    data[0, 0:3] = [x_offset + 0.0, y_offset, 0.0]
    data[0, 3:6] = [x_offset + 1.0, y_offset, 0.0]
    data[1, 0:3] = [x_offset + 1.0, y_offset, 0.0]
    data[1, 3:6] = [x_offset + 2.0, y_offset, 0.0]

    data[:, 12:15] = [1.0, 0.0, 0.0]
    data[:, 20] = 1.0
    data[:, 21] = radius

    data[0, 15] = 1
    data[1, 17] = 0
    data[0, 26] = 0
    data[1, 26] = 1

    return SimpleNamespace(data=data)


def _make_connected_tree_connection(network_id: int, x_offset: float, y_offset: float = 0.0):
    upstream = _make_tree(x_offset=x_offset, y_offset=y_offset, radius=0.1)
    downstream = _make_tree(x_offset=x_offset + 3.0, y_offset=y_offset, radius=0.1)

    vessels = [
        np.array([[[x_offset + 2.0, y_offset, 0.0, x_offset + 2.5, y_offset, 0.0, 0.1]]]),
        np.array([[[x_offset + 2.5, y_offset, 0.0, x_offset + 3.0, y_offset, 0.0, 0.1]]]),
    ]

    return SimpleNamespace(
        network_id=network_id,
        assignments=[np.array([1]), np.array([1])],
        vessels=vessels,
        connected_network=[upstream, downstream],
        forest=None,
    )


def _make_three_tree_connection(network_id: int, x_offset: float, y_offset: float = 0.0):
    tree_0 = _make_tree(x_offset=x_offset, y_offset=y_offset, radius=0.1)
    tree_1 = _make_tree(x_offset=x_offset + 3.0, y_offset=y_offset, radius=0.1)
    tree_2 = _make_tree(x_offset=x_offset + 6.0, y_offset=y_offset, radius=0.1)

    vessels = [
        np.array([[[x_offset + 2.0, y_offset, 0.0, x_offset + 2.5, y_offset, 0.0, 0.1]]]),
        np.array([[[x_offset + 2.5, y_offset, 0.0, x_offset + 3.0, y_offset, 0.0, 0.1]]]),
        np.array([[[x_offset + 5.5, y_offset, 0.0, x_offset + 6.0, y_offset, 0.0, 0.1]]]),
    ]

    return SimpleNamespace(
        network_id=network_id,
        assignments=[np.array([1]), np.array([1]), np.array([1])],
        vessels=vessels,
        connected_network=[tree_0, tree_1, tree_2],
        forest=None,
    )


def _make_unconnected_forest():
    return SimpleNamespace(
        networks=[
            [_make_tree(0.0, 0.0), _make_tree(10.0, 0.0)],
            [_make_tree(0.0, 10.0), _make_tree(10.0, 10.0)],
        ],
        connections=None,
    )


def _make_connected_forest():
    tree_connections = [
        _make_connected_tree_connection(network_id=0, x_offset=0.0, y_offset=0.0),
        _make_connected_tree_connection(network_id=1, x_offset=0.0, y_offset=5.0),
    ]
    forest = SimpleNamespace(networks=[[], []], connections=SimpleNamespace(tree_connections=tree_connections))
    for tree_connection in tree_connections:
        tree_connection.forest = forest
    return forest


def _make_three_tree_connected_forest():
    tree_connection = _make_three_tree_connection(network_id=0, x_offset=0.0, y_offset=0.0)
    forest = SimpleNamespace(
        networks=[[]],
        n_trees_per_network=[3],
        connections=SimpleNamespace(tree_connections=[tree_connection]),
    )
    tree_connection.forest = forest
    return forest


def _data_lines(path):
    lines = path.read_text(encoding="utf-8").splitlines()
    return [line for line in lines if line and not line.startswith("Vessel:")]


def _read_sidecar(path):
    sections = {"inlet": [], "outlet": []}
    current = None
    for line in path.read_text(encoding="utf-8").splitlines():
        if line in sections:
            current = line
            continue
        if line and current is not None:
            sections[current].append(tuple(float(value) for value in line.split(", ")))
    return sections


def test_tree_export_writes_requested_txt_file(tmp_path):
    tree = _make_tree(0.0, 0.0)

    written = export_spline_files(tree, tmp_path / "tree_export.txt", spline_sample_points=7)

    assert written == [tmp_path / "tree_export.txt"]
    assert written[0].is_file()
    assert sorted(path.name for path in tmp_path.glob("*.txt")) == ["tree_export.txt"]

    first_row = _data_lines(written[0])[0].split(", ")
    assert len(first_row) == 4


def test_unconnected_forest_export_writes_one_txt_per_tree(tmp_path):
    forest = _make_unconnected_forest()

    written = export_spline_files(forest, tmp_path / "forest_export.txt", spline_sample_points=5)

    expected_names = [
        "forest_export_network0_tree0.txt",
        "forest_export_network0_tree1.txt",
        "forest_export_network1_tree0.txt",
        "forest_export_network1_tree1.txt",
    ]
    assert [path.name for path in written] == expected_names
    assert not (tmp_path / "forest_export.txt").exists()
    assert sorted(path.name for path in tmp_path.glob("*.txt")) == expected_names
    assert all(path.read_text(encoding="utf-8").startswith("Vessel: 0, Number of Points: 5") for path in written)


def test_connected_forest_export_writes_one_txt_per_connected_network(tmp_path):
    forest = _make_connected_forest()

    written = export_spline_files(
        forest,
        tmp_path / "connected_export.txt",
        spline_sample_points=6,
        separate=True,
    )

    expected_names = ["connected_export_network0.txt", "connected_export_network1.txt"]
    assert [path.name for path in written] == expected_names
    assert not (tmp_path / "connected_export.txt").exists()
    assert sorted(path.name for path in tmp_path.glob("*.txt")) == expected_names

    first_row = _data_lines(written[0])[0].split(", ")
    assert len(first_row) == 5
    assert first_row[-1] in {"0", "1"}


def test_tree_export_writes_inlet_root_sidecar(tmp_path):
    tree = _make_tree(0.0, 0.0)

    written = export_spline_files(
        tree,
        tmp_path / "tree_export.txt",
        spline_sample_points=7,
        export_inlet_outlet_roots=True,
        tree_root_role="inlet",
    )

    assert written == [tmp_path / "tree_export.txt"]
    sidecar = tmp_path / "tree_export_inlet_outlet.txt"
    assert sidecar.is_file()
    assert _read_sidecar(sidecar) == {"inlet": [(0.0, 0.0, 0.0)], "outlet": []}


def test_tree_export_writes_outlet_root_sidecar(tmp_path):
    tree = _make_tree(4.0, 1.5)

    written = export_spline_files(
        tree,
        tmp_path / "tree_export.txt",
        spline_sample_points=7,
        export_inlet_outlet_roots=True,
        tree_root_role="outlet",
    )

    assert written == [tmp_path / "tree_export.txt"]
    sidecar = tmp_path / "tree_export_inlet_outlet.txt"
    assert sidecar.is_file()
    assert _read_sidecar(sidecar) == {"inlet": [], "outlet": [(4.0, 1.5, 0.0)]}


def test_tree_export_requires_root_role_for_sidecar(tmp_path):
    tree = _make_tree(0.0, 0.0)

    with pytest.raises(ValueError, match="tree_root_role"):
        export_spline_files(
            tree,
            tmp_path / "tree_export.txt",
            spline_sample_points=7,
            export_inlet_outlet_roots=True,
        )


def test_unconnected_forest_export_writes_one_sidecar_per_tree_file(tmp_path):
    forest = _make_unconnected_forest()

    written = export_spline_files(
        forest,
        tmp_path / "forest_export.txt",
        spline_sample_points=5,
        export_inlet_outlet_roots=True,
        inlet_tree_indices_by_network={0: {0}, 1: {1}},
    )

    expected = {
        "forest_export_network0_tree0_inlet_outlet.txt": {"inlet": [(0.0, 0.0, 0.0)], "outlet": []},
        "forest_export_network0_tree1_inlet_outlet.txt": {"inlet": [], "outlet": [(10.0, 0.0, 0.0)]},
        "forest_export_network1_tree0_inlet_outlet.txt": {"inlet": [], "outlet": [(0.0, 10.0, 0.0)]},
        "forest_export_network1_tree1_inlet_outlet.txt": {"inlet": [(10.0, 10.0, 0.0)], "outlet": []},
    }
    assert len(written) == 4
    for sidecar_name, expected_sections in expected.items():
        sidecar = tmp_path / sidecar_name
        assert sidecar.is_file()
        assert _read_sidecar(sidecar) == expected_sections


def test_connected_two_tree_forest_sidecar_contains_both_root_points(tmp_path):
    forest = _make_connected_forest()

    written = export_spline_files(
        forest,
        tmp_path / "connected_export.txt",
        spline_sample_points=6,
        export_inlet_outlet_roots=True,
        inlet_tree_indices_by_network={0: {0}, 1: {1}},
    )

    assert [path.name for path in written] == ["connected_export_network0.txt", "connected_export_network1.txt"]
    assert _read_sidecar(tmp_path / "connected_export_network0_inlet_outlet.txt") == {
        "inlet": [(0.0, 0.0, 0.0)],
        "outlet": [(3.0, 0.0, 0.0)],
    }
    assert _read_sidecar(tmp_path / "connected_export_network1_inlet_outlet.txt") == {
        "inlet": [(3.0, 5.0, 0.0)],
        "outlet": [(0.0, 5.0, 0.0)],
    }


def test_connected_three_tree_forest_sidecars_follow_component_tree_mapping(tmp_path):
    forest = _make_three_tree_connected_forest()

    written = export_spline_files(
        forest,
        tmp_path / "three_tree.txt",
        spline_sample_points=6,
        export_inlet_outlet_roots=True,
        inlet_tree_indices_by_network={0: {0, 2}},
    )

    assert [path.name for path in written] == ["three_tree_network0_0.txt", "three_tree_network0_1.txt"]
    assert _read_sidecar(tmp_path / "three_tree_network0_0_inlet_outlet.txt") == {
        "inlet": [(0.0, 0.0, 0.0)],
        "outlet": [(3.0, 0.0, 0.0)],
    }
    assert _read_sidecar(tmp_path / "three_tree_network0_1_inlet_outlet.txt") == {
        "inlet": [(6.0, 0.0, 0.0)],
        "outlet": [],
    }


def test_export_spline_files_rejects_invalid_inlet_tree_index(tmp_path):
    forest = _make_unconnected_forest()

    with pytest.raises(ValueError, match="Invalid inlet tree index"):
        export_spline_files(
            forest,
            tmp_path / "forest_export.txt",
            spline_sample_points=5,
            export_inlet_outlet_roots=True,
            inlet_tree_indices_by_network={0: {9}},
        )


def test_forest_spline_writer_returns_all_network_splines_and_files(tmp_path):
    all_points = [
        [
            [[0.0, 0.0, 0.0], [0.33, 0.0, 0.0], [0.66, 0.0, 0.0], [1.0, 0.0, 0.0]],
        ],
        [
            [[0.0, 1.0, 0.0], [0.33, 1.0, 0.0], [0.66, 1.0, 0.0], [1.0, 1.0, 0.0]],
        ],
    ]
    all_radii = [
        [[0.2, 0.2, 0.2, 0.2]],
        [[0.3, 0.3, 0.3, 0.3]],
    ]

    spline_functions = write_forest_splines(
        all_points,
        all_radii,
        spline_sample_points=4,
        outdir=str(tmp_path),
        write_splines=True,
    )

    assert len(spline_functions) == 2
    assert all(len(network_splines) == 1 for network_splines in spline_functions)
    assert sorted(path.name for path in tmp_path.glob("*.txt")) == [
        "network_0_b_splines.txt",
        "network_1_b_splines.txt",
    ]
